import os
import unittest
import logging
import shutil
import string
from pathlib import Path
from unittest.mock import MagicMock
import warnings

# prevent excessive warning logs
warnings.filterwarnings("ignore")
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

import tensorflow as tf
import pandas as pd
import numpy as np

from finetune import MultiLabelClassifier
from finetune.datasets import generic_download
from finetune.config import get_config

SST_FILENAME = "SST-binary.csv"


class TestMultiLabelClassifier(unittest.TestCase):
    n_sample = 20
    dataset_path = os.path.join("Data", "Classify", "SST-binary.csv")

    @classmethod
    def _download_sst(cls):
        """
        Download Stanford Sentiment Treebank to data directory
        """
        path = Path(cls.dataset_path)
        if path.exists():
            return

        path.parent.mkdir(parents=True, exist_ok=True)
        generic_download(
            url="https://s3.amazonaws.com/enso-data/SST-binary.csv",
            text_column="Text",
            target_column="Target",
            filename=SST_FILENAME,
        )

    @classmethod
    def setUpClass(cls):
        cls._download_sst()

    def setUp(self):
        self.dataset = pd.read_csv(self.dataset_path, nrows=self.n_sample * 3)
        try:
            os.mkdir("tests/saved-models")
        except FileExistsError:
            warnings.warn(
                "tests/saved-models still exists, it is possible that some test is not cleaning up properly."
            )
            pass

    def tearDown(self):
        shutil.rmtree("tests/saved-models/")

    def default_config(self, **kwargs):
        return dict(
            batch_size=2,
            max_length=128,
            n_epochs=5,
            l2_reg=0.0,
            clf_p_drop=0.0,
            **kwargs
        )

    def test_fit_predict(self):
        """
        Ensure model training does not error out
        Ensure model returns predictions of the right type
        """

        model = MultiLabelClassifier(**self.default_config())
        train_sample = self.dataset.sample(n=self.n_sample)
        valid_sample = self.dataset.sample(n=self.n_sample)
        model.fit(train_sample.Text, [[t, 6, 3] for t in train_sample.Target])

        probabilities = model.predict_proba(valid_sample.Text)
        for proba in probabilities:
            self.assertIsInstance(proba, dict)

        predictions = model.predict(valid_sample.Text)
        for prediction in predictions:
            self.assertIsInstance(prediction[0], (str, np.int, np.int64))
            self.assertIn(3, prediction)
            self.assertIn(6, prediction)




import logging

import numpy as np

from sklearn.model_selection import train_test_split

from finetune import Classifier
from finetune.datasets.stanford_sentiment_treebank import StanfordSentimentTreebank
from finetune.base_models import TCN
logging.basicConfig(level=logging.DEBUG)

if __name__ == "__main__":
    # Train and evaluate on SST
    dataset = StanfordSentimentTreebank(nrows=1000).dataframe
    model = Classifier(
        # val_size=0.,
        max_length=64,
        base_model=TCN,
        # early_stopping_steps=100,
        n_epochs=5, 
        keep_best_model=True,
        debugging_logs=True
    )
    trainX, testX, trainY, testY = train_test_split(dataset.Text.values, dataset.Target.values, test_size=0.3, random_state=42)
    model.fit(trainX, trainY)
    accuracy = np.mean(model.predict(testX) == testY)
    print('Test Accuracy: {:0.2f}'.format(accuracy))



"""
Convert plain text to format accepted by model (token idxs + special tokens).
"""
import warnings
import functools
from collections import namedtuple

import spacy
import tensorflow as tf


NLP = spacy.load("en", disable=["parser", "tagger", "ner", "textcat"])

EncodedOutput = namedtuple(
    "EncodedOutput",
    [
        "token_ids",  # list of list of subtoken ids (ints)
        "tokens",  # list of list of subtokens (strs)
        "labels",  # list of list of labels
        "char_locs",  # list of list of character locations (ints)
        "char_starts",  # list of list of character starts (locs are character ends) (ints)
    ],
)
EncodedOutput.__new__.__defaults__ = (None,) * len(EncodedOutput._fields)
ArrayEncodedOutput = namedtuple(
    "ArrayEncodedOutput",
    [
        "token_ids",  # int array shape (batch, seq_length)
        "tokens",  # list of list of subtokens (str) passed through from `EncoderOutput`
        "labels",  # object array shape (batch, seq_length)
        "char_locs",  # list of list of char_locs (int) passed through from `EncoderOutput`
        "mask",  # int array shape (batch, seq_length)
    ],
)
ArrayEncodedOutput.__new__.__defaults__ = (None,) * len(ArrayEncodedOutput._fields)

SUBS = {"—": "-", "–": "-", "―": "-", "…": "...", "´": "'"}


def get_pairs(word):
    """
    Return set of symbol pairs in a word.
    word is represented as tuple of symbols (symbols being variable-length strings)
    """
    pairs = set()
    prev_char = word[0]
    for char in word[1:]:
        pairs.add((prev_char, char))
        prev_char = char
    return pairs


def _flatten(nested_lists):
    return functools.reduce(lambda x, y: x + y, nested_lists, [])


class BaseEncoder(object):
    """
    Base class for text encoders.
    Translates raw texts into structured token arrays
    """

    UNK_IDX = 0

    def __init__(self, encoder_path, vocab_path):
        self.initialized = False
        self.encoder_path = encoder_path
        self.vocab_path = vocab_path

        # Required public attributes -- consider refactor to prevent requiring direct
        # access of these attributes
        self.special_tokens = None
        self.start_token = None
        self.delimiter_token = None
        self.end_token = None
        self.encoder = None
        self.decoder = None

    def _lazy_init(self):
        pass

    @property
    def vocab_size(self):
        self._lazy_init()
        return len(self.encoder)

    def __getitem__(self, key):
        return self.encoder[key]

    def __setitem__(self, key, value):
        self.encoder[key] = value

    def _encode(self, texts, labels=None):
        """
        Convert a batch of raw text to a batch of byte-pair encoded token indices.
        """
        raise NotImplementedError

    def decode(self, ids):
        """
        Convert a batch of ids [batch_size, id] into text(ish).
        """

        raise NotImplementedError

    def _cut_and_concat(
        self,
        *,
        encoded,
        max_length,
        special_tokens=None,
        start=None,
        delimiter=None,
        end=None
    ):
        """
        Takes some tokenized text and arranges it into a format that maximises the amount of kept text from each
        whilst keeping the overall sequence length within max_length tokens. It also adds the 3 special tokens. Start,
         Classify and Delimiter.
        :param encoded: Lists of shape [batch, n_fields, num_tokens]
        :param max_length: Int representing the max length of a single sample
        :param start: Override the default start token.
        :param delimiter: Override the default delimiter token.
        :param end: Override the default classify token
        :return: Formatted outputs of the form. [batch, num_tokens] where num_tokens' <= max_length
        """
        start = start or special_tokens or self.start_token
        delimiter = delimiter or special_tokens or self.delimiter_token
        clf_token = end or special_tokens or self.end_token

        num_samples = len(encoded)
        adjusted_max_length = max_length - num_samples - 1
        allocated_max_len = adjusted_max_length // num_samples

        overflows = [allocated_max_len - len(sequence) for sequence in encoded]
        spare = sum(overflows)

        if spare >= 0:
            cut_len = None
        else:
            warnings.warn(
                "Document is longer than max length allowed, trimming document to {} tokens.".format(
                    max_length
                )
            )
            empty_tokens = sum(max(overflow, 0) for overflow in overflows)
            num_over = [max(overflow, 0) for overflow in overflows].count(0)
            if num_over == 0:
                cut_len = allocated_max_len
            else:
                cut_len = allocated_max_len + (empty_tokens // num_over)

        joined = [start]
        for d in encoded:
            joined += d[:cut_len] + [delimiter]
        joined = joined[:-1] + [clf_token]

        return joined

    def _token_length(self, token):
        return len(token)

    def encode_multi_input(self, Xs, Y=None, max_length=None, pad_token=None):
        """
        Encodes the text for passing to the model, also tracks the location of each token to allow reconstruction.
        It can also, optionally, construct a per-token labels as required for training.
        :param Xs: A list of lists of string -- [n_fields, n_segments]
        :param Y: A list of list of targets -- [n_batch, n_segments]
        :param max_length: Max length of the sequences.
        :return: A Labeled Sequence Object.
        """
        token_ids = []
        tokens = []
        positions = []
        labels = []

        # for each field in that example
        for field in Xs:
            assert isinstance(
                field, (list, tuple)
            ), "This should be a list of strings, instead it's {} {}".format(
                tf.contrib.framework.nest.map_structure(type, field), Xs
            )
            encoded = self._encode(field, labels=Y)
            token_ids.append(_flatten(encoded.token_ids))
            tokens.append(_flatten(encoded.tokens))
            positions.append(_flatten(encoded.char_locs))
            labels.append(_flatten(encoded.labels))
            if len(tokens[-1]) > (max_length - 2):
                warnings.warn(
                    "Some examples are longer than the max_length. Please trim documents or increase `max_length`. "
                    "Fallback behaviour is to use the first {} byte-pair encoded tokens".format(
                        max_length - 2
                    )
                )

        # merge fields + truncate if necessary
        token_ids = self._cut_and_concat(encoded=token_ids, max_length=max_length)
        tokens = self._cut_and_concat(encoded=tokens, max_length=max_length)
        locations = self._cut_and_concat(
            encoded=positions, max_length=max_length, special_tokens=-1
        )

        if Y is None:
            labels = None
        else:
            labels = self._cut_and_concat(
                encoded=labels, max_length=max_length, special_tokens=pad_token
            )

        return EncodedOutput(
            token_ids=token_ids,
            tokens=tokens,
            labels=labels,
            char_locs=locations,
        )

    def __setstate__(self, state):
        self.__init__()

    def __getstate__(self):
        return {"Encoder": None}



import os
import unittest
import shutil
import json
import random
import time
from pathlib import Path
import warnings

# prevent excessive warning logs
warnings.filterwarnings('ignore')
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import pandas as pd
import numpy as np
from sklearn.model_selection import train_test_split

from finetune.base_models import BERTModelCased, GPT2Model, RoBERTa
from finetune import Classifier, DeploymentModel, SequenceLabeler, ComparisonRegressor
from finetune.base_models import GPTModel
from finetune.datasets import generic_download

SST_FILENAME = "SST-binary.csv"


class TestDeploymentModel(unittest.TestCase):
    n_sample = 20
    do_comparison = True
    base_model = GPTModel

    classifier_path = "tests/saved-models/deployment_classifier.jl"
    comparison_regressor_path = "tests/saved-models/deployment_comparison_regressor.jl"
    sequence_labeler_path = "tests/saved-models/deployment_sequence_labeler.jl"
    classifier_dataset_path = os.path.join(
        'Data', 'Classify', 'SST-binary.csv'
    )

    @classmethod
    def _download_data(cls):
        # Download Stanford Sentiment Treebank to data directory

        path = Path(cls.classifier_dataset_path)
        if path.exists():
            return

        path.parent.mkdir(parents=True, exist_ok=True)
        generic_download(
            url="https://s3.amazonaws.com/enso-data/SST-binary.csv",
            text_column="Text",
            target_column="Target",
            filename=SST_FILENAME
        )

    @classmethod
    def setUpClass(cls):
        cls._download_data()
        
        # dataset preparation
        cls.classifier_dataset = pd.read_csv(cls.classifier_dataset_path, nrows=cls.n_sample * 10)

        path = os.path.join(os.path.dirname(__file__), "data", "testdata.json")
        with open(path, 'rt') as fp:
            cls.texts, cls.labels = json.load(fp)

        cls.animals = ["dog", "cat", "horse", "cow", "pig", "sheep", "goat", "chicken", "guinea pig", "donkey", "turkey", "duck", "camel", "goose", "llama", "rabbit", "fox"]
        cls.numbers = ["one", "two", "three", "four", "five", "six", "seven", "eight", "nine", "ten", "eleven", "twelve", "thirteen", "fourteen", "fifteen", "sixteen"]
        
        # train and save sequence labeler for later use
        cls.s = SequenceLabeler(**cls.default_seq_config())
        cls.s.fit(cls.texts * 10, cls.labels * 10)
        cls.s.save(cls.sequence_labeler_path)
        
        # train and save classifier for later use
        train_sample = cls.classifier_dataset.sample(n=cls.n_sample*10)
        cls.cl = Classifier(**cls.default_config())
        cls.cl.fit(train_sample.Text, train_sample.Target)
        cls.cl.save(cls.classifier_path)

        if cls.do_comparison:
            # train and save comparison regressor for use
            cls.cr = ComparisonRegressor()
    
            n_per = 150
            similar = []
            different = []
            for dataset in [cls.animals, cls.numbers]:
                for i in range(n_per // 2):
                    similar.append([random.choice(dataset), random.choice(dataset)])
            for i in range(n_per):
                different.append([random.choice(cls.animals), random.choice(cls.numbers)])

            targets = np.asarray([1] * len(similar) + [0] * len(different))
            data = similar + different

            cls.x_tr, cls.x_te, cls.t_tr, cls.t_te = train_test_split(data, targets, test_size=0.3, random_state=42)

            cls.cr = ComparisonRegressor(**cls.default_config())
            cls.cr.fit(cls.x_tr, cls.t_tr)
            cls.cr.save(cls.comparison_regressor_path)

    def setUp(self):
        try:
            os.mkdir("tests/saved-models")
        except FileExistsError:
            warnings.warn("tests/saved-models still exists, it is possible that some test is not cleaning up properly.")

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree("tests/saved-models/")

    @classmethod
    def default_config(cls, **kwargs):
        defaults = {
            'batch_size': 2,
            'max_length': 256,
            'n_epochs': 3,
            'adapter_size': 64,
            'chunk_long_sequences': False,
            'base_model': cls.base_model
        }
        defaults.update(kwargs)
        return defaults

    @classmethod
    def default_seq_config(cls, **kwargs):
        d = dict(
            batch_size=2,
            max_length=256,
            lm_loss_coef=0.0,
            val_size=0,
            adapter_size=64,
            base_model=cls.base_model,
            interpolate_pos_embed=False,
        )
        d.update(**kwargs)
        return d

    def test_cached_predict(self):
        """
        Ensure second call to predict is faster than first, including loading.
        """
        model = DeploymentModel(featurizer=self.base_model, **self.default_config())
        model.load_featurizer()

        valid_sample = self.classifier_dataset.sample(n=self.n_sample)
        start = time.time()
        model.load_custom_model(self.classifier_path)
        model.predict(valid_sample.Text[:1].values)
        first = time.time()
        model.predict(valid_sample.Text[:1].values)
        second = time.time()

        first_prediction_time = (first - start)
        second_prediction_time = (second - first)
        self.assertLess(second_prediction_time, first_prediction_time)
        model.close()

    def test_switching_models(self):
        """
        Ensure model can switch out weights without erroring out
        """
        model = DeploymentModel(featurizer=self.base_model, **self.default_config())
        model.load_featurizer()
        # test transitioning from any of [sequence labeling, comparison, default] to any other
        model.load_custom_model(self.classifier_path)
        if self.do_comparison:
            model.load_custom_model(self.comparison_regressor_path)
        model.load_custom_model(self.sequence_labeler_path)
        model.load_custom_model(self.classifier_path)
        model.load_custom_model(self.sequence_labeler_path)
        if self.do_comparison:
            model.load_custom_model(self.comparison_regressor_path)
        model.load_custom_model(self.classifier_path)
        model.close()

    def test_reasonable_predictions(self):
        """
        Ensure model produces reasonable predictions after loading weights
        """
        model = DeploymentModel(featurizer=self.base_model, **self.default_seq_config())
        model.load_featurizer()
        
        # test same output as weights loaded with Classifier model
        valid_sample = self.classifier_dataset.sample(n=self.n_sample)
        model.load_custom_model(self.classifier_path)
        deployment_preds = model.predict_proba(valid_sample.Text.values)
        model.close()
        classifier_preds = self.cl.predict_proba(valid_sample.Text.values)
        for c_pred, d_pred in zip(classifier_preds, deployment_preds):
            self.assertTrue(list(c_pred.keys()) == list(d_pred.keys()))
            for c_pred_val, d_pred_val in zip(c_pred.values(), d_pred.values()):
                np.testing.assert_almost_equal(c_pred_val, d_pred_val, decimal=4)
        
        if self.do_comparison:
            # test same output as weights loaded with Comparison Regressor model
            model = DeploymentModel(featurizer=self.base_model, **self.default_seq_config())
            model.load_featurizer()
            model.load_custom_model(self.comparison_regressor_path)
            deployment_preds = model.predict(self.x_te)
            model.close()
            compregressor = ComparisonRegressor.load(self.comparison_regressor_path)
            compregressor_preds = compregressor.predict(self.x_te)
            for c_pred, d_pred in zip(compregressor_preds, deployment_preds):
                np.testing.assert_almost_equal(c_pred, d_pred, decimal=4)

    def test_large_predict(self):
        """
        Ensure model does not have OOM issues with large inputs for inference
        """
        large_dataset = self.animals*100
        model = DeploymentModel(featurizer=self.base_model, **self.default_config())
        model.load_featurizer()
        model.load_custom_model(self.classifier_path)
        model.predict(large_dataset)
        model.close()

    def test_fast_switch(self):
        """
        Ensure model can load/reload weights and predict in reasonable time
        """
        model = DeploymentModel(featurizer=self.base_model, **self.default_config())
        model.load_featurizer()
        model.load_custom_model(self.classifier_path)
        model.predict(['finetune'])

        start = time.time()
        model.load_custom_model(self.sequence_labeler_path)
        model.predict(['finetune sequence'])
        end = time.time()
        self.assertGreater(2.5, end - start)

        if self.do_comparison:
            start = time.time()
            model.load_custom_model(self.comparison_regressor_path)
            model.predict([['finetune', 'compare']])
            end = time.time()
            self.assertGreater(2.5, end - start)
        model.close()


class TestDeploymentBert(TestDeploymentModel):
    base_model = BERTModelCased


class TestDeploymentGPT(TestDeploymentModel):
    base_model = GPTModel


class TestDeploymentGPT2(TestDeploymentModel):
    base_model = GPT2Model


class TestDeploymentRoberta(TestDeploymentModel):
    base_model = RoBERTa


"""
Convert plain text to format accepted by model (token idxs + special tokens).
"""
import re
import json
import os
import warnings
import codecs

import numpy as np
from ftfy.fixes import uncurl_quotes

import finetune
from finetune.encoding.input_encoder import NLP, EncodedOutput, BaseEncoder, get_pairs


FINETUNE_FOLDER = os.path.dirname(finetune.__file__)
ENCODER_PATH = os.path.join(FINETUNE_FOLDER, "model", "gpt", "encoder.json")
VOCAB_PATH = os.path.join(FINETUNE_FOLDER, "model", "gpt", "vocab.bpe")
SUBS = {"—": "-", "–": "-", "―": "-", "…": "...", "´": "'"}


def _text_standardize(text):
    """
    Fixes some issues the spacy tokenizer had on books corpus
    Also handles whitespace standardization
    """
    text = re.sub(
        """(-+|~+|!+|"+|;+|\?+|\++|,+|\)+|\(+|\\+|\/+|\*+|\[+|\]+|}+|{+|\|+|_+)""",
        r" \1 ",
        text,
    )
    text = re.sub("\s*\n\s*", " \n ", text)
    text = re.sub("[^\S\n]+", " ", text)
    return uncurl_quotes(text.strip().lower())


class GPTEncoder(BaseEncoder):
    """
    A modified wrapper for a public python BPE tokenizer. The modifications allow encoding directly into the formats
    required for finetune. Particularly with respect to formatting with multiple inputs.
    """

    UNK_IDX = 0

    def __init__(self, encoder_path=ENCODER_PATH, vocab_path=VOCAB_PATH):
        super().__init__(encoder_path=encoder_path, vocab_path=vocab_path)

    def _lazy_init(self):
        # Must set
        if self.initialized:
            return

        with open(self.encoder_path, "r") as f:
            self.encoder = json.load(f)

        with codecs.open(self.vocab_path, encoding="utf8"):
            merges = (
                codecs.open(self.vocab_path, encoding="utf8").read().split("\n")[1:-1]
            )
            merges = [tuple(merge.split()) for merge in merges]
            self.bpe_ranks = dict(zip(merges, range(len(merges))))

        self.decoder = {v: k for k, v in self.encoder.items()}

        self.special_tokens = ["_start_", "_delimiter_", "_classify_"]
        for token in self.special_tokens:
            self.encoder[token] = len(self.encoder)

        self.decoder = {v: k for k, v in self.encoder.items()}
        self.cache = {}
        self.start_token = self.encoder["_start_"]
        self.delimiter_token = self.encoder["_delimiter_"]
        self.end_token = self.encoder["_classify_"]
        self.initialized = True

    def _token_length(self, token):
        return len(token.strip().replace("</w>", ""))

    def bpe(self, token):
        word = tuple(token[:-1]) + (token[-1] + "</w>",)
        if token in self.cache:
            return self.cache[token]
        pairs = get_pairs(word)

        if not pairs:
            return token + "</w>"

        while True:
            bigram = min(pairs, key=lambda pair: self.bpe_ranks.get(pair, float("inf")))
            if bigram not in self.bpe_ranks:
                break
            first, second = bigram
            new_word = []
            i = 0
            while i < len(word):
                try:
                    j = word.index(first, i)
                    new_word.extend(word[i:j])
                    i = j
                except:
                    new_word.extend(word[i:])
                    break

                if word[i] == first and i < len(word) - 1 and word[i + 1] == second:
                    new_word.append(first + second)
                    i += 2
                else:
                    new_word.append(word[i])
                    i += 1
            new_word = tuple(new_word)
            word = new_word
            if len(word) == 1:
                break
            else:
                pairs = get_pairs(word)
        word = " ".join(word)
        if word == "\n  </w>":
            word = "\n</w>"
        self.cache[token] = word
        return word

    def _encode(self, texts, labels=None):
        """
        Convert a batch of raw text to a batch of byte-pair encoded token indices.
        """
        self._lazy_init()
        batch_tokens = []
        batch_token_idxs = []
        batch_label_idxs = []
        batch_char_ends = (
            []
        )  # to account for the fact that some BPEs have different lengths than their original tokens (e.g. special characters such as bullets)
        batch_char_starts = []
        label = None
        skipped = 0
        for i, text in enumerate(texts):
            if labels is not None:
                label = labels[i]

            raw_text = text.lower()
            
            # Only fine to apply this fix because it preserves character locations
            ftfy_text = uncurl_quotes(raw_text)
            tokens = NLP(_text_standardize(text))
            if not tokens:
                skipped += 1
                continue
            i -= skipped
            subtokens = []
            subtoken_idxs = []
            char_starts = []
            char_ends = []
            token_start = 0

            for j, token in enumerate(tokens):
                bpe_toks = self.bpe(token.text).split(" ")

                try:
                    if token.text.strip():
                        token_start = ftfy_text.index((token.text.strip()), token_start)
                except ValueError:
                    warnings.warn(
                        "Failed to find token `{}` in text.".format(token.text)
                    )
                    continue

                subtokens.extend(bpe_toks)
                subtoken_idxs.extend(
                    [self.encoder.get(SUBS.get(t, t), self.UNK_IDX) for t in bpe_toks]
                )

                assert len("".join(bpe_toks).replace("</w>", "")) == len(
                    token.text.replace(" ", "")
                )

                if np.sum([len(tok.replace("</w>", "")) for tok in bpe_toks]) > len(
                    token
                ):  # the BPEs comprising a token are longer than the token itself
                    token_char_ends = (
                        np.asarray([len(token.text.strip()) for tok in bpe_toks])
                        + token_start
                    )
                else:
                    token_char_ends = (
                        np.cumsum([len(tok.replace("</w>", "")) for tok in bpe_toks])
                        + token_start
                    )
                
                token_char_starts = [token_start] + token_char_ends[:-1].tolist()
                token_start += len(token.text.strip())
                char_ends.extend(token_char_ends)
                char_starts.extend(token_char_starts)

            batch_tokens.append(subtokens)
            batch_token_idxs.append(subtoken_idxs)
            batch_char_ends.append(char_ends)
            batch_char_starts.append(char_starts)
            if labels is not None:
                batch_label_idxs.append([label] * len(subtoken_idxs))

        return EncodedOutput(
            token_ids=batch_token_idxs,
            tokens=batch_tokens,
            labels=batch_label_idxs,
            char_locs=batch_char_ends,
            char_starts=batch_char_starts,
        )

    def decode(self, ids):
        """
        Convert a batch of ids [batch_size, id] into text(ish).
        """

        return "".join(
            [self.decoder.get(word_idx, "<unk>") for word_idx in ids]
        ).replace("</w>", " ")


def aggregate_to_full_tokens(attn, tokens, token_starts, token_ends, attention=True):
    to_combine = []
    spacy_attn = []
    spacy_token_starts = []
    spacy_token_ends = []
    spacy_start = None
    for token, prob, start, end in zip(tokens, attn, token_starts, token_ends):
        to_combine.append(prob)
        if spacy_start is None:
            spacy_start = start
        if token.endswith("</w>"):
            if attention:
                spacy_attn.append(np.max(to_combine, 0))
            else:
                spacy_attn.append(to_combine[-1])
            spacy_token_starts.append(spacy_start)
            spacy_token_ends.append(end)
            to_combine = []
            spacy_start = None

    if attention:
        spacy_attn = spacy_attn / sum(spacy_attn)
        key = "attention_weights"
    else:
        key = "explanation"

    return {
        key: spacy_attn,
        "token_starts": spacy_token_starts,
        "token_ends": spacy_token_ends,
    }


def finetune_to_indico_explain(raw_texts, attn_weights, encoder, attention=True):
    """
    Maps the attention weights or explain probability values one-to-one with the raw text tokens.
    
    :param raw_texts: A list of segmented text of the form list(list(str))
    :param attn_weights: An array of attention weights of shape [batch, seq_len]
    :param encoder: The encoder used in the model that output the attention weights
    :param attention: A boolean flag, true for inputting attention matrices, false for explain probabilities.
    :return: A list of dictionaries, each with the following keys:
    - 'attention_weights': A list of attention weights for each token in the text
    - 'token_starts': A list of the start tokens for each spacy token in the text
    - 'token_ends': A list of the end tokens for each spacy token in the text

    """
    spacy_outputs = []

    encoded_docs = encoder._encode(raw_texts)

    for doc_idx, (raw_text) in enumerate(raw_texts):
        tokens = encoded_docs.tokens[doc_idx]
        token_ends = encoded_docs.char_locs[doc_idx]
        token_lengths = [encoder._token_length(token) for token in tokens]
        token_starts = [end - length for end, length in zip(token_ends, token_lengths)]
        if attention:
            # offset of one to take into account the start token
            clf_token_idx = len(tokens) + 1
            # take the average values over the attention heads for the attention weights of the classify token
            attn = np.mean(attn_weights[doc_idx], axis=0)[clf_token_idx][
                1:clf_token_idx
            ]  # [num_tokens]
        else:
            # There is no start token
            clf_token_idx = len(tokens)
            attn = attn_weights[doc_idx][:clf_token_idx]
        # map one-to-one with spacy tokenization
        spacy_output = aggregate_to_full_tokens(
            attn, tokens, token_starts, token_ends, attention=attention
        )
        spacy_outputs.append(spacy_output)
    return spacy_outputs






from finetune.base import BaseModel, PredictMode

def process_long_sequence(X, task, probas=False)
    tasks = ['sequence_labeling', 'classification', 'regression']
    assert task in tasks, 'invalid task for processing long sequences'
    chunk_size = self.config.max_length - 2
    step_size = chunk_size // 3
    if task == 'sequence_labeling':
        arr_encoded = list(itertools.chain.from_iterable(self.input_pipeline._text_to_ids([x]) for x in X))
    else:
        arr_encoded = list(itertools.chain.from_iterable(self.input_pipeline._text_to_ids(x) for x in X))
       
    labels, batch_probas = [], []
    pred_keys = [PredictMode.NORMAL]
    if task != 'regression':
        pred_keys.append(PredictMode.PROBAS)
    for pred in self._inference(X, predict_keys=[PredictMode.PROBAS, PredictMode.NORMAL], n_examples=len(arr_encoded)):
        labels.append(self.input_pipeline.label_encoder.inverse_transform(pred[PredictMode.NORMAL]))
        if task != 'regression':
            batch_probas.append(pred[PredictMode.PROBAS])

    all_subseqs = []
    all_labels = []
    all_probs = []
    all_positions = []

    doc_idx = -1
    for chunk_idx, (label_seq, proba_seq) in enumerate(zip(labels, batch_probas)):
        position_seq = arr_encoded[chunk_idx].char_locs
        start_of_doc = arr_encoded[chunk_idx].token_ids[0][0] == self.input_pipeline.text_encoder.start
        end_of_doc = (
                chunk_idx + 1 >= len(arr_encoded) or
                arr_encoded[chunk_idx + 1].token_ids[0][0] == self.input_pipeline.text_encoder.start
        )
        """
        Chunk idx for prediction.  Dividers at `step_size` increments.
        [  1  |  1  |  2  |  3  |  3  ]
        """
        start, end = 0, None
        if start_of_doc:
            # if this is the first chunk in a document, start accumulating from scratch
            doc_subseqs = []
            doc_labels = []
            doc_probs = []
            doc_positions = []
            doc_starts = []

            doc_idx += 1
            start_of_token = 0
            if not end_of_doc:
                end = step_size * 2
        else:
            if end_of_doc:
                # predict on the rest of sequence
                start = step_size
            else:
                # predict only on middle third
                start, end = step_size, step_size * 2

        if task == 'sequence_labeling':
            label_seq = label_seq[start:end]
            position_seq = position_seq[start:end]
            proba_seq = proba_seq[start:end]

            for label, position, proba in zip(label_seq, position_seq, proba_seq):
                if position == -1:
                    # indicates padding / special tokens
                    continue
                
                # if there are no current subsequence
                # or the current subsequence has the wrong label
                if not doc_subseqs or label != doc_labels[-1] or per_token:
                    # start new subsequence
                    doc_subseqs.append(X[doc_idx][start_of_token:position])
                    doc_labels.append(label)
                    doc_probs.append([proba])
                    doc_positions.append((start_of_token, position))
                    doc_starts.append(start_of_token)
                else:
                    # continue appending to current subsequence
                    doc_subseqs[-1] = X[doc_idx][doc_starts[-1]:position]
                    doc_probs[-1].append(proba)
                start_of_token = position

            if end_of_doc:
                # last chunk in a document
                prob_dicts = []
                for prob_seq in doc_probs:
                    # format probabilities as dictionary
                    probs = np.mean(np.vstack(prob_seq), axis=0)
                    prob_dicts.append(dict(zip(self.input_pipeline.label_encoder.classes_, probs)))
                    if self.multi_label:
                        del prob_dicts[-1][self.config.pad_token]

                all_subseqs.append(doc_subseqs)
                all_labels.append(doc_labels)
                all_probs.append(prob_dicts)
                all_positions.append(doc_positions)

        _, doc_annotations = finetune_to_indico_sequence(raw_texts=X, subseqs=all_subseqs, labels=all_labels,
                                                            probs=all_probs, none_value=self.config.pad_token,
                                                            subtoken_predictions=self.config.subtoken_predictions)

        if per_token:
            return [
                {
                    'tokens': _spacy_token_predictions(
                        raw_text=raw_text,
                        tokens=tokens,
                        probas=probas, 
                        positions=positions
                    ),
                    'prediction': predictions,
                }
                for raw_text, tokens, labels, probas, positions, predictions in zip(
                    X, all_subseqs, all_labels, all_probs, all_positions, doc_annotations
                )
            ]
        else:
            return doc_annotations
    else:
        doc_labels.append(label)
        doc_probs.append(proba)

        if task == 'classification':
                if end_of_doc:
                # last chunk in a document
                mean_pool = np.mean(doc_probs, axis=0)
                pred = np.argmax(mean_pool)
                one_hot = np.zeros_like(mean_pool)
                one_hot[pred] = 1
                label = self.input_pipeline.label_encoder.inverse_transform([one_hot])
                all_labels.append(label)
                all_probs.append(mean_pool)

            if probas:
                return all_probs
            else:
                return all_labels
        else:
                if end_of_doc:
                    mean_pool = np.mean(doc_probs, axis=0)
                    predictions = self.input_pipeline.label_encoder.inverse_transform([mean_pool])
                    all_labels.append(label)
            




import os
import unittest
import warnings

# prevent excessive warning logs
warnings.filterwarnings('ignore')
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import random
import numpy as np

from sklearn.model_selection import train_test_split
from finetune import ComparisonOrdinalRegressor, OrdinalRegressor

ANIMALS = ["dog", "cat", "horse", "cow", "pig", "sheep", "goat", "chicken", "guinea pig", "donkey", "turkey", "duck", "camel", "goose", "llama", "rabbit", "fox"]
NUMBERS = ["one", "two", "three", "four", "five", "six", "seven", "eight", "nine", "ten", "eleven", "twelve", "thirteen", "fourteen", "fifteen", "sixteen"]


class TestOrdinal(unittest.TestCase):
    n_sample = 100

    def default_config(self, **kwargs):
        d = dict(
            batch_size=2,
            max_length=16,
            n_epochs=5,
            val_size=0.,
            l2_reg=0.,
            interpolate_pos_embed=False,
            low_memory_mode=True,
        )
        d.update(kwargs)
        return d

    def setUp(self):
        random.seed(42)
        np.random.seed(42)

    
    def test_reasonable_predictions(self):
        """
        Ensure model training does not error out
        Ensure model returns predictions of the right type
        Test model loss at least outperforms naive baseline
        """
        model = OrdinalRegressor(**self.default_config())

        # fake dataset generation
        targets = np.asarray([1] * len(ANIMALS) + [0] * len(NUMBERS))
        data = ANIMALS + NUMBERS

        x_tr, x_te, t_tr, t_te = train_test_split(data, targets, test_size=0.3)
        model.finetune(x_tr, t_tr)

        predictions = model.predict(x_te)
        mse = np.mean([(pred - true)**2 for pred, true in zip(predictions, t_te)])
        naive_baseline = max(np.mean(targets == 1), np.mean(targets == 0))
        naive_baseline_mse = np.mean([(naive_baseline - true)**2 for true in t_te])
        self.assertIsInstance(predictions, list)
        self.assertGreater(naive_baseline_mse, mse)
    

    def test_reasonable_predictions_comparison(self):
        """
        Ensure model training does not error out
        Ensure model returns predictions of the right type
        Test model loss at least outperforms naive baseline
        """
        model = ComparisonOrdinalRegressor(**self.default_config())

        n_per = 100
        similar = []
        different = []
        for dataset in [ANIMALS, NUMBERS]:
            for i in range(n_per // 2):
                similar.append([random.choice(dataset), random.choice(dataset)])
        for i in range(n_per):
            different.append([random.choice(ANIMALS), random.choice(NUMBERS)])

        targets = np.asarray([1] * len(similar) + [0] * len(different))
        data = similar + different

        x_tr, x_te, t_tr, t_te = train_test_split(data, targets, test_size=0.3)
        model.finetune(x_tr, t_tr)

        predictions = model.predict(x_te)
        mse = np.mean([(pred - true)**2 for pred, true in zip(predictions, t_te)])
        naive_baseline = max(np.mean(targets == 1), np.mean(targets == 0))
        naive_baseline_mse = np.mean([(naive_baseline - true)**2 for true in t_te])
        self.assertIsInstance(predictions, list)
        self.assertGreater(naive_baseline_mse, mse)

    def test_reasonable_predictions_unshared_weights(self):
        """
        Ensure model training does not error out
        Ensure model returns predictions of the right type
        Does not analyze model loss since unshared weights perform poorly at
        these low data volumes
        """
        model = OrdinalRegressor(**self.default_config())

        # fake dataset generation

        targets = np.asarray([1] * len(ANIMALS) + [0] * len(NUMBERS))
        data = ANIMALS + NUMBERS

        x_tr, x_te, t_tr, t_te = train_test_split(data, targets, test_size=0.3)
        model.finetune(x_tr, t_tr)

        predictions = model.predict(x_te)
        self.assertIsInstance(predictions, list)

if __name__ == '__main__':
    unittest.main()



import itertools
import copy
import tensorflow as tf
import numpy as np
from imblearn.over_sampling import RandomOverSampler
from sklearn.utils import shuffle

from finetune.base import BaseModel
from finetune.encoding.target_encoders import OneHotLabelEncoder
from finetune.nn.target_blocks import classifier
from finetune.input_pipeline import BasePipeline
from finetune.model import PredictMode
from finetune.base_models.gpt.encoder import finetune_to_indico_explain


class ClassificationPipeline(BasePipeline):
    def resampling(self, Xs, Y):
        if self.config.oversample:
            idxs, Ys = shuffle(
                *RandomOverSampler().fit_sample([[i] for i in range(len(Xs))], Y)
            )
            return [Xs[i[0]] for i in idxs], Ys
        return Xs, Y

    def _target_encoder(self):
        return OneHotLabelEncoder()

    def feed_shape_type_def(self):
        TS = tf.TensorShape
        return (
            ({"tokens": tf.int32, "mask": tf.float32}, tf.float32),
            (
                {
                    "tokens": TS([self.config.max_length, 2]),
                    "mask": TS([self.config.max_length]),
                },
                TS([self.target_dim]),
            ),
        )


class Classifier(BaseModel):
    """ 
    Classifies a single document into 1 of N categories.

    :param config: A :py:class:`finetune.config.Settings` object or None (for default config).
    :param \**kwargs: key-value pairs of config items to override.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _get_input_pipeline(self):
        return ClassificationPipeline(self.config)

    def featurize(self, X):
        """
        Embeds inputs in learned feature space. Can be called before or after calling :meth:`finetune`.

        :param X: list or array of text to embed.
        :returns: np.array of features of shape (n_examples, embedding_size).
        """
        return super().featurize(X)

    def predict(self, X, probas=False):
        """
        Produces a list of most likely class labels as determined by the fine-tuned model.

        :param X: list or array of text to embed.
        :returns: list of class labels.
        
        
        Chunk idx for prediction.  Dividers at `step_size` increments.
        [  1  |  1  |  2  |  3  |  3  ]
        """
        all_labels = []
        all_probs = []
        doc_probs = []
        for _, start_of_doc, end_of_doc, _, proba in self.process_long_sequence(X):
            start, end = 0, None
            doc_probs.append(proba)

            if end_of_doc:
                # last chunk in a document
                mean_pool = np.mean(doc_probs, axis=0)
                pred = np.argmax(mean_pool)
                one_hot = np.zeros_like(mean_pool)
                one_hot[pred] = 1
                label = self.input_pipeline.label_encoder.inverse_transform([one_hot])
                label = np.squeeze(label).tolist()
                all_labels.append(label)
                all_probs.append(mean_pool)
                doc_probs = []

        if probas:
            return all_probs
        else:
            assert len(all_labels) == len(X)
            return np.asarray(all_labels)

    def predict_proba(self, X):
        """
        Produces a probability distribution over classes for each example in X.

        :param X: list or array of text to embed.
        :returns: list of dictionaries.  Each dictionary maps from a class label to its assigned class probability.
        """
        # this is simply a vector of probabilities, not a dict from classes to class probs
        raw_probas = self.predict(X, probas=True)
        classes = self.input_pipeline.label_encoder.classes_
        formatted_predictions = []
        for probas in raw_probas:
            formatted_predictions.append(dict(zip(classes, probas)))
        return formatted_predictions

    def finetune(self, X, Y=None, batch_size=None):
        """
        :param X: list or array of text.
        :param Y: integer or string-valued class labels.
        :param batch_size: integer number of examples per batch. When N_GPUS > 1, this number
                           corresponds to the number of training examples provided to each GPU.
        """
        return super().finetune(X, Y=Y, batch_size=batch_size)

    @classmethod
    def get_eval_fn(cls):
        return lambda labels, targets: np.mean(
            np.asarray(labels) == np.asarray(targets)
        )

    @staticmethod
    def _target_model(
        config, featurizer_state, targets, n_outputs, train=False, reuse=None, **kwargs
    ):
        if "explain_out" in featurizer_state:
            shape = tf.shape(featurizer_state["explain_out"])  # batch, seq, hidden
            flat_explain = tf.reshape(
                featurizer_state["explain_out"], [shape[0] * shape[1], shape[2]]
            )
            hidden = tf.concat((featurizer_state["features"], flat_explain), 0)
        else:
            hidden = featurizer_state["features"]

        clf_out = classifier(
            hidden=hidden,
            targets=targets,
            n_targets=n_outputs,
            config=config,
            train=train,
            reuse=reuse,
            **kwargs
        )
        if "explain_out" in featurizer_state:
            logits = clf_out["logits"]
            clf_out["logits"] = logits[: shape[0]]
            clf_out["explanation"] = tf.nn.softmax(
                tf.reshape(
                    logits[shape[0] :],
                    tf.concat((shape[:2], [tf.shape(logits)[-1]]), 0),
                ),
                -1,
            )
        return clf_out

    def explain(self, Xs):
        explanation = self._inference(
            Xs, predict_keys=[PredictMode.EXPLAIN, PredictMode.NORMAL]
        )
        classes = self.input_pipeline.label_encoder.target_labels
        out = []
        bases = []
        preds = []
        for values in explanation:
            explain_sample = values[PredictMode.EXPLAIN]
            preds.append(values[PredictMode.NORMAL])
            out.append(explain_sample[1:])
            bases.append(explain_sample[0])
        processed = finetune_to_indico_explain(
            Xs, out, self.input_pipeline.text_encoder, attention=False
        )

        for base, sample, cls in zip(bases, processed, preds):
            weights = sample["explanation"]
            weights = np.array([base] + weights[:-1]) - weights
            n_classes = weights.shape[-1]
            norm = (
                np.max([np.abs(np.max(weights, 0)), abs(np.min(weights, 0))], 0)
                * n_classes
            )
            explanation = weights / norm + 1 / n_classes

            sample["explanation"] = {
                c: explanation[:, i] for i, c in enumerate(classes)
            }
            sample["prediction"] = self.input_pipeline.label_encoder.inverse_transform(
                [cls]
            )[0]

        return processed

    def _predict_op(self, logits, **kwargs):
        return tf.contrib.seq2seq.hardmax(logits)

    def _predict_proba_op(self, logits, **kwargs):
        return tf.nn.softmax(logits, -1)



import logging
import numpy as np

from sklearn.model_selection import train_test_split

import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
import matplotlib.colors

from finetune import Classifier
from finetune.datasets.stanford_sentiment_treebank import StanfordSentimentTreebank
from finetune.base_models.gpt.model import GPTModel

logging.basicConfig(level=logging.DEBUG)


def sentences_heatmap(sentences, probs, clf):
    fig, axes = plt.subplots(len(sentences), 1, figsize=(20, len(sentences) * 0.95), dpi=160, squeeze=False)
    axes = axes[:, 0]

    axes[0].axis("off")

    for i, ax in enumerate(axes):
        ax.add_patch(Rectangle((0.0, 0.05), 0.99, 0.90, fill=None, alpha=1,
                               color="r" if clf[i] == 0 else "g", linewidth=2))

        word_pos = 0.00
        cmap = plt.cm.copper
        norm = matplotlib.colors.Normalize(vmin=0.0, vmax=1.0)

        for j, (word, prob) in enumerate(zip(sentences[i], probs[i])):
            ax.text(word_pos, 0.5, word,
                    horizontalalignment='left',
                    verticalalignment='center',
                    fontsize=8, color=cmap(norm(prob)),
                    transform=ax.transAxes, fontweight=700)
            word_pos += .0036 * (len(word) + 1)  # to move the word for the next iter
        ax.axis('off')
    plt.subplots_adjust(wspace=0, hspace=0)
    plt.tight_layout()
    plt.savefig("test_sentences.png")


if __name__ == "__main__":
    dataset = StanfordSentimentTreebank(nrows=1000).dataframe
    model = Classifier(
        interpolate_pos_embed=False,
        n_epochs=1,
        batch_size=2,
        lr_warmup=0.1,
        val_size=0.0,
        max_length=64,
        prefit_init=False,
        base_model=GPTModel
    )

    trainX, testX, trainY, testY = train_test_split(dataset.Text.values, dataset.Target.values, test_size=0.3,
                                                    random_state=42)
    model.fit(trainX, trainY)
    samples = model.explain(testX)
    words = []
    probs = []
    clfs = []

    for sample, text in zip(samples, testX):
        sample_words = []
        sample_probs = []
        for s, e, p in zip(sample["token_starts"], sample["token_ends"], sample['explanation'][1]):
            sample_words.append(text[s:e])
            sample_probs.append(p)
        print(list(zip(sample_words, sample_probs)), text)
        words.append(sample_words)
        probs.append(sample_probs)
        clfs.append(sample["prediction"])
    sentences_heatmap(words, probs, clfs)
    accuracy = np.mean(testY == [s["prediction"] for s in samples])
    print('Test Accuracy: {:0.2f}'.format(accuracy))



import os
from concurrent.futures import ThreadPoolExecutor
import logging
import sys

import joblib
import numpy as np
import tensorflow as tf
from tensorflow.contrib.estimator.python.estimator.early_stopping import _StopOnPredicateHook
from tensorflow.estimator import SessionRunHook


from finetune.errors import FinetuneError
from finetune.config import get_config
from finetune.util.metrics import read_eval_metrics


LOGGER = logging.getLogger("finetune")


class SaverHook(_StopOnPredicateHook):
    def __init__(
        self,
        saver,
        estimator,
        keep_best_model,
        early_stopping_steps,
        steps_per_epoch,
        eval_frequency,
    ):
        super().__init__(
            self.stop_if_no_metric_improvement_fn,
            run_every_secs=None,
            run_every_steps=eval_frequency,
        )
        self.get_current_weights = False
        self.included = None
        self.saver = saver
        self.keep_best_model = keep_best_model
        self.early_stopping_steps = early_stopping_steps or sys.maxsize
        self.steps_per_epoch = steps_per_epoch
        self.estimator = estimator

    def stop_if_no_metric_improvement_fn(self):
        if not self.keep_best_model:
            return False
        eval_results = read_eval_metrics(self.estimator.eval_dir())
        if len(eval_results) == 0:
            return False
        most_recent_eval = max(eval_results.items(), key=lambda x: x[0])  # last steps.
        best_eval = min(eval_results.items(), key=lambda x: x[1]["loss"])  # lowest_loss
        if most_recent_eval == best_eval:
            self.get_current_weights = True
        steps_diff = most_recent_eval[0] - best_eval[0]
        tf.logging.info("No improvement in {} steps".format(steps_diff))

        if (
            steps_diff > self.early_stopping_steps
            and most_recent_eval[0] > self.steps_per_epoch
        ):
            LOGGER.info(
                "Early stopping triggered.".format(
                    steps_diff
                )
            )
            return True
        return False

    def begin(self):
        super().begin()
        self.included = tf.global_variables()

    def after_run(self, run_context, run_values):
        super().after_run(run_context, run_values)
        if self.get_current_weights:
            self.saver.variables = dict(
                zip(
                    (var.name for var in self.included),
                    run_context.session.run(self.included),
                )
            )
            self.get_current_weights = False

    def end(self, session):
        self.stop_if_no_metric_improvement_fn()
        if (
            not self.keep_best_model
            or self.saver.variables is None
            or self.get_current_weights
        ):
            self.saver.variables = dict(
                zip((var.name for var in self.included), session.run(self.included))
            )


class InitializeHook(SessionRunHook):
    def __init__(self, saver, model_portion="entire_model"):
        self.saver = saver
        self.model_portion = model_portion
        self.need_to_refresh = True  # between predicts of the same model
        self.refresh_base_model = (
            False
        )  # after we have loaded a model that has trained the entire featurizer, we need to reload from fallback for the next model
        self.init_fn = self.saver.get_scaffold_init_fn()

    def after_create_session(self, session, coord):
        if self.model_portion != "entire_model" and self.need_to_refresh:
            if self.model_portion == "target":
                self.init_fn(None, session, self.model_portion)
            else:
                self.init_fn(
                    None, session, "whole_featurizer"
                )  # after_create_session only called at load_featurizer in deployment_model, so load entire featurizer
            self.need_to_refresh = False
        elif self.model_portion == "entire_model":
            self.init_fn(None, session, self.model_portion)

    def before_run(self, run_context):
        if "featurizer" in self.model_portion and (
            self.need_to_refresh or self.refresh_base_model
        ):
            if self.model_portion == "whole_featurizer":
                self.refresh_base_model = True
            self.init_fn(
                None, run_context.session, self.model_portion, self.refresh_base_model
            )
            self.need_to_refresh = False
            self.refresh_base_model = False


class BatchedVarLoad:
    """
    Basic idea pulled from variable.load. Possible this will change in 2.X so worth tracking what
    that fn changes too and mirror it here.
    """
    def __init__(self):
        self.ops = []
        self.feed = dict()

    def run(self, session):
        session.run(self.ops, feed_dict=self.feed)
        self.ops = []
        self.feed = dict()

    def add(self, var, val):
        if hasattr(var, "_values"):
            underlying_vars = var._values
        else:
            underlying_vars = [var]
        for v in underlying_vars:
            if v.initializer not in self.ops:
                self.ops.append(v.initializer)
            self.feed[v.initializer.inputs[1]] = val


class Saver:
    def __init__(
        self,
        fallback_filename=None,
        exclude_matches=None,
        variable_transforms=None,
        save_dtype=None,
    ):
        self.variable_transforms = variable_transforms or []
        self.exclude_matches = exclude_matches
        self.variables = None
        self.save_dtype = save_dtype
        if fallback_filename is not None:
            self.set_fallback(fallback_filename)

    def set_fallback(self, fallback_filename):
        self.tpe = ThreadPoolExecutor()
        if not os.path.exists(fallback_filename):
            raise FileNotFoundError("Error loading base model - file not found.")
        self.fallback_filename = fallback_filename
        self.fallback_future = self.tpe.submit(joblib.load, fallback_filename)
        self.fallback_ = None

    @property
    def fallback(self):
        if self.fallback_ is None:
            self.fallback_ = self.fallback_future.result()
            self.fallback_future = None
            self.tpe.shutdown()
        return self.fallback_

    def get_saver_hook(
        self,
        estimator,
        keep_best_model,
        steps_per_epoch,
        early_stopping_steps,
        eval_frequency,
    ):
        return SaverHook(
            self,
            estimator=estimator,
            keep_best_model=keep_best_model,
            steps_per_epoch=steps_per_epoch,
            early_stopping_steps=early_stopping_steps,
            eval_frequency=eval_frequency,
        )

    def save(self, finetune_obj, path, mkdir=True):
        if self.variables is None:
            raise FinetuneError("Cowardly refusing to save default model.")
        if self.exclude_matches is not None:
            variables = {
                k: v for k, v in self.variables.items() if self.exclude_matches not in k
            }
        else:
            variables = self.variables

        names, values = variables.keys(), variables.values()
        if isinstance(path, str):
            folder = os.path.dirname(path)
            os.makedirs(folder, exist_ok=True)
        if self.save_dtype is not None:
            LOGGER.info("Saving with {} precision.".format(self.save_dtype.__name__))
            values = [a.astype(self.save_dtype) for a in values]

        var_names_reduced, vals_reduced = self.remove_unchanged(
            names, values, self.fallback
        )
        var_dict = dict(zip(var_names_reduced, vals_reduced))
        assert len(vals_reduced) == len(var_names_reduced) == len(var_dict)
        joblib.dump((var_dict, finetune_obj), path)

    def load(self, path):
        self.variables, finetune_obj = joblib.load(path)
        finetune_obj.config = get_config(
            error_on_invalid_keywords=False, 
            **dict(finetune_obj.config)
        )
        return finetune_obj

    def get_scaffold_init_fn(self):
        def init_fn(scaffold, session, model_portion=None, refresh_base_model=False):
            var_loader = BatchedVarLoad()
            self.var_val = []
            if self.variables is not None:
                variables_sv = self.variables
            else:
                variables_sv = dict()
            all_vars = tf.global_variables()
            zero_out_adapters = False
            if (
                model_portion != "entire_model"
            ):  # we must be loading in the case of two separate estimators
                all_vars, zero_out_adapters = self.subset_to_load(
                    model_portion, refresh_base_model, all_vars
                )

            for var in all_vars:
                name = var.name
                saved_var = None
                if name in variables_sv.keys():
                    saved_var = variables_sv[name]
                elif name in self.fallback.keys():
                    saved_var = self.fallback[name]

                if zero_out_adapters and "adapter" in name:
                    var_loader.add(var, np.zeros(var.get_shape().as_list()))
                if saved_var is not None:
                    for func in self.variable_transforms:
                        saved_var = func(name, saved_var)
                    var_loader.add(var, saved_var)
            var_loader.run(session)
        return init_fn

    def subset_to_load(self, model_portion, refresh_base_model, all_vars):
        assert model_portion in [
            "featurizer",
            "target",
            "whole_featurizer",
        ], "Must be using separate estimators if loading before graph creation"
        base = [v for v in all_vars if "target" not in v.name]
        zero_out_adapters = False
        if (
            model_portion == "whole_featurizer"
        ):  # load every weight in featurizer - used to initialize and for loading without adapters
            to_load = base
            adapters = [v for v in base if "adapter" in v.name]
            zero_out_adapters = True
        elif (
            model_portion == "featurizer"
        ):  # update featurizer, loading adapters and scaling weights
            norm_variable_scopes = ["b:0", "g:0", "beta:0", "gamma:0"]
            to_load = (
                base
                if refresh_base_model
                else [
                    v
                    for v in base
                    if "target" not in v.name
                    and (
                        "adapter" in v.name
                        or any(scope in v.name for scope in norm_variable_scopes)
                    )
                ]
            )
        elif model_portion == "target":  # update target model weights
            to_load = [v for v in all_vars if "target" in v.name]
        return to_load, zero_out_adapters

    def remove_unchanged(self, variable_names, variable_values, fallback_vars):
        skips = []
        for var_val, var_name in zip(variable_values, variable_names):
            skip = False
            for fb_var_name, fb_var in fallback_vars.items():
                if fb_var_name == var_name:
                    for func in self.variable_transforms:
                        fb_var = func(var_name, fb_var)
                    if np.allclose(fb_var, var_val):
                        skip = True
                        break
            skips.append(skip)
        return (
            [var for skip, var in zip(skips, variable_names) if not skip],
            [var_val for skip, var_val in zip(skips, variable_values) if not skip],
        )




import numpy as np
import tensorflow as tf
import copy

from finetune.errors import FinetuneError
from finetune.base import BaseModel
from finetune.target_models.classifier import Classifier, ClassificationPipeline
from finetune.encoding.input_encoder import ArrayEncodedOutput


class ComparisonPipeline(ClassificationPipeline):
    def _format_for_encoding(self, X):
        return [X]

    def _text_to_ids(self, pair, Y=None, pad_token=None):
        """
        Format comparison examples as a list of IDs

        pairs: Array of text, shape [batch, 2]
        """
        assert (
            self.config.chunk_long_sequences is False
        ), "Chunk Long Sequences is not compatible with comparison"
        arr_forward = next(super()._text_to_ids(pair, Y=None))
        reversed_pair = pair[::-1]
        arr_backward = next(super()._text_to_ids(reversed_pair, Y=None))
        kwargs = arr_forward._asdict()
        kwargs["tokens"] = [arr_forward.tokens, arr_backward.tokens]
        kwargs["token_ids"] = np.stack(
            [arr_forward.token_ids, arr_backward.token_ids], 0
        )
        kwargs["mask"] = np.stack([arr_forward.mask, arr_backward.mask], 0)
        yield ArrayEncodedOutput(**kwargs)

    def feed_shape_type_def(self):
        TS = tf.TensorShape
        return (
            ({"tokens": tf.int32, "mask": tf.float32}, tf.float32),
            (
                {
                    "tokens": TS([2, self.config.max_length, 2]),
                    "mask": TS([None, self.config.max_length]),
                },
                TS([self.target_dim]),
            ),
        )


class Comparison(Classifier):
    """
    Compares two documents to solve a classification task.

    :param config: A :py:class:`finetune.config.Settings` object or None (for default config).
    :param \**kwargs: key-value pairs of config items to override.
    """

    defaults = {"chunk_long_sequences": False}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.config.chunk_long_sequences:
            raise FinetuneError(
                "Multifield model is incompatible with chunk_long_sequences = True in config."
            )

    def _get_input_pipeline(self):
        return ComparisonPipeline(self.config)

    @staticmethod
    def _target_model(
        config,
        *,
        featurizer_state,
        targets,
        n_outputs,
        train=False,
        reuse=None,
        **kwargs
    ):
        featurizer_state = featurizer_state.copy()
        featurizer_state["sequence_features"] = tf.abs(
            tf.reduce_sum(featurizer_state["sequence_features"], 1)
        )
        featurizer_state["features"] = tf.abs(
            tf.reduce_sum(featurizer_state["features"], 1)
        )
        return Classifier._target_model(
            config,
            featurizer_state=featurizer_state,
            targets=targets,
            n_outputs=n_outputs,
            train=train,
            reuse=reuse,
            **kwargs
        )

    def predict(self, pairs):
        """
        Produces a list of most likely class labels as determined by the fine-tuned model.


        :param pairs: Array of text, shape [batch, 2]
        :returns: list of class labels.
        """
        return BaseModel.predict(self, pairs)

    def predict_proba(self, pairs):
        """
        Produces a probability distribution over classes for each example in X.


        :param pairs: Array of text, shape [batch, 2]
        :returns: list of dictionaries.  Each dictionary maps from a class label to its assigned class probability.
        """
        return BaseModel.predict_proba(self, pairs)

    def featurize(self, pairs):
        """
        Embeds inputs in learned feature space. Can be called before or after calling :meth:`finetune`.

        :param pairs: Array of text, shape [batch, 2]
        :returns: np.array of features of shape (n_examples, embedding_size).
        """
        return BaseModel.featurize(self, pairs)



import warnings
import copy
from collections import defaultdict

import numpy as np

from finetune.util.logging import truncate_text
from finetune.encoding.input_encoder import NLP
from finetune.errors import FinetuneError


def assign_associations(associations, none_value, idx_lookup):
    candidates = dict()
    for association in associations:
        for bpe_idx, candidate_idx, candidate_label, candidate_prob in association:
            if candidate_label == none_value:
                continue
            if idx_lookup[bpe_idx] not in candidates:
                candidates[idx_lookup[bpe_idx]] = []
            candidates[idx_lookup[bpe_idx]].append(
                (idx_lookup[candidate_idx], candidate_label, candidate_prob)
            )

    # TODO some how sample these candidates eg maximum probabilities, to fit some schema
    candidates = {
        k: max(v, key=lambda x: x[2]) for k, v in candidates.items()
    }  # for now just pick maximum prob
    return candidates


def _merge_confidences(annotation):
    """
    Collapse list of confidences down to a single mean confidence.
    """
    if not "confidence" in annotation or not len(annotation["confidence"]):
        return

    labels = annotation["confidence"][0].keys()
    annotation["confidence"] = {
        label: np.mean(
            [confidences[label] for confidences in annotation["confidence"]]
        ).tolist()
        for label in labels
    }


def round_to_nearest_start_and_end(label, token_starts, token_ends, text):
    # Update label start / end / text to align with nearest token start_token and end
    # Applies in-place modification to `label` obj.
    start_distances = np.abs(np.asarray(token_starts) - label["start"])
    end_distances = np.abs(np.asarray(token_ends) - label["end"])
    label["start"] = token_starts[np.argmin(start_distances)]
    label["end"] = token_ends[np.argmin(end_distances)]
    label["text"] = text[label["start"] : label["end"]]


def finetune_to_indico_sequence(
    raw_texts,
    subseqs,
    labels,
    probs=None,
    none_value=None,
    subtoken_predictions=False,
    associations=None,
):
    """
    Maps from the labeled substring format into the 'indico' format. This is the exact inverse operation to
    :meth indico_to_finetune_sequence:.

    The indico format is as follows:
        Raw text for X,
        Labels as a list of dicts, with each dict in the form:
        {
            'start': <Character index of the start_token of the labeled sequence>,
            'end': <Character index of the end of the labeled sequence>,
            'label': <A categorical label (int or string) that represents the category of the subsequence,
            'text': <Optionally, a field with the subsequence contained between the start and end.
        }

    The Labeled substring, or finetune internal, format is as follows.
    Each item of the data is a list strings of the form:
        ["The quick brown", "fox", "jumped over the lazy", ...]
    With the corresponding labels:
        ["PAD", "animal", "PAD", ...]

    It is the :param none_value: that is used to populate the PAD labels.
    :param data: A list of segmented text of the form list(list(str))
    :param labels: Categorical labels for each sub-string in data.
    :param none_value: The none value used to encode the input format.
    :return: Texts, annoatations both in the 'indico' format.
    """
    annotations = []
    loop_vals = zip(raw_texts, subseqs, labels, probs or [None] * len(raw_texts))
    for doc_idx, (raw_text, doc_seq, label_seq, prob_seq) in enumerate(loop_vals):
        spacy_tokens = NLP(raw_text)
        spacy_token_starts = [token.idx for token in spacy_tokens]
        spacy_token_ends = [token.idx + len(token.text) for token in spacy_tokens]
        doc_annotations = []
        annotation_ranges = set()
        raw_annotation_start = 0
        subtoken_to_label_idx = []
        for i, (sub_str, raw_label, confidences) in enumerate(
            zip(doc_seq, label_seq, prob_seq or [None] * len(doc_seq))
        ):
            subtoken_to_label_idx.append(len(doc_annotations))
            if not isinstance(raw_label, tuple):
                label_list = [raw_label]
            else:
                label_list = raw_label

            for label_idx, label in enumerate(label_list):
                stripped_text = sub_str.strip()

                if subtoken_predictions:
                    raw_annotation_start = raw_text.find(sub_str, raw_annotation_start)
                    raw_annotation_end = raw_annotation_start + len(sub_str)
                else:
                    raw_annotation_start = raw_text.find(
                        stripped_text, raw_annotation_start
                    )
                    raw_annotation_end = raw_annotation_start + len(stripped_text)

                if raw_annotation_start == -1:
                    warnings.warn(
                        "Failed to find predicted sequence: {} in text: {}.".format(
                            sub_str, raw_text
                        )
                    )
                    continue

                extended_existing_label = False
                for item in doc_annotations:
                    # handle case where we extend existing annotation
                    if (
                        # same label
                        item["label"] == label
                        # and only separated by whitespace
                        and item["end"] <= raw_annotation_end
                        and not raw_text[item["end"] : raw_annotation_start].strip()
                    ):
                        item["end"] = raw_annotation_end
                        item["text"] = raw_text[item["start"] : raw_annotation_end]
                        if "confidence" in item and confidences is not None:
                            item["confidence"].append(confidences)
                        extended_existing_label = True
                        break

                if extended_existing_label or label == none_value:
                    continue

                annotation_start, annotation_end = (
                    int(raw_annotation_start),
                    int(raw_annotation_end),
                )

                annotation = {
                    "start": int(annotation_start),
                    "end": int(annotation_end),
                    "label": label,
                    "text": raw_text[annotation_start:annotation_end],
                }

                # if we don't want to allow subtoken predictions, adjust start and end to match
                # the start and ends of the nearest full tokens
                if not subtoken_predictions:
                    round_to_nearest_start_and_end(
                        annotation, spacy_token_starts, spacy_token_ends, raw_text
                    )

                if confidences is not None:
                    annotation["confidence"] = [confidences]

                if annotation["start"] >= annotation["end"]:
                    continue

                # prevent duplicate annotation edge case
                annotation_tuple = (annotation["start"], annotation["end"], label)
                if annotation_tuple not in annotation_ranges:
                    annotation_ranges.add(annotation_tuple)
                    doc_annotations.append(annotation)

        if associations:
            associations_seq = assign_associations(
                associations[doc_idx], none_value, subtoken_to_label_idx
            )
            for label_i, annotation in enumerate(doc_annotations):
                if label_i in associations_seq:
                    index, relationship, prob = associations_seq[label_i]
                    annotation["associations"] = {
                        "index": index,
                        "relationship": relationship,
                        "prob": prob,
                    }

        doc_annotations = sorted(
            [dict(items) for items in doc_annotations], key=lambda x: span(x)
        )

        for annotation in doc_annotations:
            _merge_confidences(annotation)

        annotations.append(doc_annotations)
    return raw_texts, annotations


def span(annotation):
    # represents annotation by its char start and end idx
    return (annotation["start"], annotation["end"])


def sorted_insert(annotations, annotation):
    # iterate in reverse order until you find the proper location to insert the new annotation
    idx = 0
    n = len(annotations)
    while idx < len(annotations) and span(annotation) < span(annotations[n - idx - 1]):
        idx += 1

    # mutate annotations list in place
    annotations.insert(n - idx, annotation)


def overlap(current_annotation, annotation):
    return (
        current_annotation["start"] < annotation["end"] <= current_annotation["end"]
    ) or (annotation["start"] < current_annotation["end"] <= annotation["end"])


def overlap_handler(current_annotation, annotation, text, multi_label):
    """
    Scenarios:
        <> --> current_annotation
        [] --> annotation
        
    1) < [ > ]
    2) [ < > ]
    3) < [ ] >
    """
    if current_annotation["start"] <= annotation["start"]:
        first, second = current_annotation, annotation
    else:
        first, second = annotation, current_annotation

    final_delimiter = min(first["end"], second["end"])
    final_label = second["label"] if (second["end"] > first["end"]) else first["label"]
    overlapping_text = text[second["start"] : final_delimiter]
    end = max(first["end"], second["end"])

    first_chunk = {
        "start": first["start"],
        "end": second["start"],
        "label": first["label"],
        "text": text[first["start"] : second["start"]],
    }

    if multi_label:
        second_label = first["label"] | second["label"]
    else:
        if first["label"] != second["label"] and (len(overlapping_text.strip()) > 1):
            warnings.warn(
                "Found overlapping annotations: {} and {}. \n"
                "Consider setting `multi_label_sequences` to `True` in your config.".format(
                    annotation, current_annotation
                )
            )
        spacy_tokens = NLP(text)
        spacy_token_starts = [token.idx for token in spacy_tokens]
        if second["label"] in spacy_token_starts:
            second_label = second["label"]
        elif final_delimiter in spacy_token_starts:
            second_label = first["label"]
        else:
            second_label = first["label"]

    second_chunk = {
        "start": second["start"],
        "end": final_delimiter,
        "label": second_label,
        "text": overlapping_text,
    }

    third_chunk = {
        "start": final_delimiter,
        "end": end,
        "label": final_label,
        "text": text[final_delimiter:end],
    }
    chunks = [first_chunk, second_chunk, third_chunk]
    chunks = [c for c in chunks if c["start"] != c["end"]]
    return chunks


def indico_to_finetune_sequence(
    texts, labels=None, encoder=None, multi_label=True, none_value=None
):
    """
    Maps from the 'indico' format sequence labeling data. Into a labeled substring format. This is the exact inverse of
    :meth finetune_to_indico_sequence:.

    The indico format is as follows:
        Raw text for X,
        Labels as a list of dicts, with each dict in the form:
         labeled sequence>,
            'end': <Character index of the end of the labeled sequence>,
            'label': <A categorical label (int or string) that represents the category of the subsequence,
            'text': <A field containing the sub-sequence contained between the start and end.
        }

    The Labeled substring, or finetune internal, format is as follows.
    Each item of the data is a list strings of the form:{
            'start': <Character index of the start_token of the
        ["The quick brown", "fox", "jumped over the lazy", ...]
    With the corresponding labels:
        ["PAD", "animal", "PAD", ...]

    It is the :param none_value: that is used to populate the PAD labels.

    :param texts: A list of raw text.
    :param labels: A list of targets of the form list(list(dict))).
    :param none_value: A categorical label to use as the none value.
    :return: Segmented Text, Labels of the form described above.
    """
    all_subseqs = []
    all_labels = []
    all_association_idx = []
    all_association_type = []
    all_idxs = []

    # placeholder for inference time
    if labels is None:
        labels = [[]] * len(texts)

    encoded_docs = encoder._encode(texts)

    labels = copy.deepcopy(labels)

    for doc_idx, (text, label_seq) in enumerate(zip(texts, labels)):
        tokens = encoded_docs.tokens[doc_idx]
        token_ends = encoded_docs.char_locs[doc_idx]
        token_starts = encoded_docs.char_starts[doc_idx]
        label_seq = sorted(label_seq, key=lambda x: x["start"])
        merged_annotations = []

        doc_association_idx = []
        doc_association_type = []
        doc_current_label_idx = []

        for i, label in enumerate(label_seq):
            # Add the index to track for annotation
            label["idx"] = i
            # Check user hasn't accidentally mislabeled something
            if (
                label.get("text") is not None
                and label["text"] != text[label["start"] : label["end"]]
            ):

                raise ValueError(
                    "Annotation text does not match text specified by `start` and `end` indexes. "
                    "Text provided: `{}`.  Text extracted: `{}`.".format(
                        label.get("text"), text[label["start"] : label["end"]]
                    )
                )

        queue = sorted(label_seq, key=lambda x: (x["start"], x["end"]))

        for label in queue:
            label["label"] = {label["label"]}
            round_to_nearest_start_and_end(label, token_starts, token_ends, text)

        while len(queue):
            current_annotation = queue.pop(0)
            # for each existing merged annotation
            for annotation in merged_annotations:
                # no overlap possible, check next merged annotation
                if annotation["end"] <= current_annotation["start"]:
                    continue

                # no overlap possible and no possibility of future overlap because of sorting
                # append and move on to next item in queue
                if annotation["start"] > current_annotation["end"]:
                    sorted_insert(merged_annotations, current_annotation)
                    break

                # if the merged annotation overlaps, remove it and break it up
                # into it's component parts.  process each component individually
                elif overlap(current_annotation, annotation):
                    merged_annotations.remove(annotation)
                    split_annotations = overlap_handler(
                        current_annotation, annotation, text, multi_label
                    )
                    queue = split_annotations + queue
                    break
            else:
                # annotations can only be added to the list of merged annotations once all
                # of their conflicts have already been resolved
                sorted_insert(merged_annotations, current_annotation)

        for annotation in merged_annotations:
            annotation["label"] = tuple(annotation["label"])
        # Add none labels
        current_idx = 0
        all_annotations = []
        for annotation in merged_annotations:
            if annotation["start"] > current_idx:
                # Add none span
                all_annotations.append(
                    {
                        "start": current_idx,
                        "end": annotation["start"],
                        "text": text[current_idx : annotation["start"]],
                        "label": tuple([none_value]),
                    }
                )
            # Copy over labeled span
            all_annotations.append(annotation)
            current_idx = annotation["end"]

        # Add span for the rest of the document
        try:
            # End is start of remaining span
            last_chunk_end_idx = max(
                [annotation["end"] for annotation in all_annotations]
            )
        except ValueError:
            # No labels, entire document is one span
            last_chunk_end_idx = 0

        if last_chunk_end_idx != len(text):
            all_annotations.append(
                {
                    "start": last_chunk_end_idx,
                    "end": len(text),
                    "text": text[last_chunk_end_idx : len(text)],
                    "label": tuple([none_value]),
                }
            )

        if not multi_label:
            # if `multi_label_sequences` is False, flatten labels
            for annotation in all_annotations:
                assert len(annotation["label"]) == 1
                annotation["label"] = annotation["label"][0]

        for annotation in all_annotations:
            if "association" not in annotation:
                doc_association_idx.append(-1)
                doc_association_type.append(none_value)
                doc_current_label_idx.append(-2)
            else:
                doc_association_idx.append(annotation["association"]["index"])
                doc_association_type.append(annotation["association"]["relationship"])
                doc_current_label_idx.append(annotation["idx"])

        doc_subseqs = [annotation["text"] for annotation in all_annotations]
        doc_labels = [annotation["label"] for annotation in all_annotations]

        all_subseqs.append(doc_subseqs)
        all_labels.append(doc_labels)
        all_association_idx.append(doc_association_idx)
        all_association_type.append(doc_association_type)
        all_idxs.append(doc_current_label_idx)

    return all_subseqs, all_labels, all_association_type, all_association_idx, all_idxs



import os
from urllib.parse import urljoin

from finetune.base_models import SourceModel
from finetune.base_models.bert.encoder import (
    BERTEncoder,
    BERTEncoderMultuilingal,
    BERTEncoderLarge,
    DistilBERTEncoder
)

from finetune.base_models.bert.roberta_encoder import RoBERTaEncoder
from finetune.base_models.bert.featurizer import bert_featurizer
from finetune.util.download import BERT_BASE_URL, GPT2_BASE_URL, ROBERTA_BASE_URL, FINETUNE_BASE_FOLDER


class BERTModelCased(SourceModel):
    is_bidirectional = True
    encoder = BERTEncoder
    featurizer = bert_featurizer
    settings = {
        "n_embed": 768,
        "n_epochs": 8,
        "n_heads": 12,
        "n_layer": 12,
        "act_fn": "gelu",
        "lr_warmup": 0.1,
        "lr": 1e-5,
        "l2_reg": 0.01,
        "bert_intermediate_size": 3072,
        "base_model_path": os.path.join("bert", "bert_small_cased.jl"),
    }
    required_files = [
        {
            "file": os.path.join(FINETUNE_BASE_FOLDER, "model", "bert", filename),
            "url": urljoin(BERT_BASE_URL, filename),
        }
        for filename in ["bert_small_cased.jl", "vocab.txt"]
    ]


class BERTModelLargeCased(SourceModel):
    is_bidirectional = True
    encoder = BERTEncoderLarge
    featurizer = bert_featurizer
    settings = {
        "n_embed": 1024,
        "n_epochs": 8,
        "n_heads": 16,
        "n_layer": 24,
        "num_layers_trained": 24,
        "act_fn": "gelu",
        "lr_warmup": 0.1,
        "lr": 1e-5,
        "l2_reg": 0.01,
        "bert_intermediate_size": 4096,
        "base_model_path": os.path.join("bert", "bert_large_cased.jl"),
    }
    required_files = [
        {
            "file": os.path.join(FINETUNE_BASE_FOLDER, "model", "bert", filename),
            "url": urljoin(BERT_BASE_URL, filename),
        }
        for filename in ["bert_large_cased.jl", "vocab_large.txt"]
    ]


class BERTModelLargeWWMCased(SourceModel):
    is_bidirectional = True
    encoder = BERTEncoderLarge
    featurizer = bert_featurizer
    settings = {
        "n_embed": 1024,
        "n_epochs": 8,
        "n_heads": 16,
        "n_layer": 24,
        "num_layers_trained": 24,
        "act_fn": "gelu",
        "lr_warmup": 0.1,
        "lr": 1e-5,
        "l2_reg": 0.01,
        "bert_intermediate_size": 4096,
        "base_model_path": os.path.join("bert", "bert_wwm_large_cased.jl"),
    }
    required_files = [
        {
            "file": os.path.join(FINETUNE_BASE_FOLDER, "model", "bert", filename),
            "url": urljoin(BERT_BASE_URL, filename),
        }
        for filename in ["bert_wwm_large_cased.jl", "vocab_large.txt"]
    ]


class BERTModelMultilingualCased(SourceModel):
    is_bidirectional = True
    encoder = BERTEncoderMultuilingal
    featurizer = bert_featurizer
    settings = {
        "n_embed": 768,
        "n_epochs": 8,
        "n_heads": 12,
        "n_layer": 12,
        "act_fn": "gelu",
        "lr_warmup": 0.1,
        "lr": 1e-5,
        "l2_reg": 0.01,
        "bert_intermediate_size": 3072,
        "base_model_path": os.path.join("bert", "bert_small_multi_cased.jl"),
    }
    required_files = [
        {
            "file": os.path.join(FINETUNE_BASE_FOLDER, "model", "bert", filename),
            "url": urljoin(BERT_BASE_URL, filename),
        }
        for filename in ["bert_small_multi_cased.jl", "vocab_multi.txt"]
    ]


class RoBERTa(SourceModel):
    is_bidirectional = True
    encoder = RoBERTaEncoder
    featurizer = bert_featurizer
    settings = {
        "n_embed": 768,
        "n_epochs": 8,
        "n_heads": 12,
        "n_layer": 12,
        "act_fn": "gelu",
        "lr_warmup": 0.1,
        "lr": 1e-5,
        "l2_reg": 0.1,
        "epsilon": 1e-8,
        "bert_intermediate_size": 3072,
        "bert_use_pooler": False,
        "max_length": 512,
        "base_model_path": os.path.join("bert", "roberta-model-sm.jl"),
    }
    required_files = [
        {
            "file": os.path.join(
                FINETUNE_BASE_FOLDER, "model", "bert", "roberta-model-sm.jl"
            ),
            "url": urljoin(ROBERTA_BASE_URL, "roberta-model-sm.jl"),
        },
        {
            "file": os.path.join(FINETUNE_BASE_FOLDER, "model", "bert", "dict.txt"),
            "url": urljoin(ROBERTA_BASE_URL, "dict.txt"),
        },
        {
            "file": os.path.join(FINETUNE_BASE_FOLDER, "model", "gpt2", "vocab.bpe"),
            "url": urljoin(GPT2_BASE_URL, "vocab.bpe"),
        },
        {
            "file": os.path.join(FINETUNE_BASE_FOLDER, "model", "gpt2", "encoder.json"),
            "url": urljoin(GPT2_BASE_URL, "encoder.json"),
        },
    ]

class RoBERTaLarge(SourceModel):
    is_bidirectional = True
    encoder = RoBERTaEncoder
    featurizer = bert_featurizer
    settings = {
        "n_embed": 1024,
        "n_epochs": 8,
        "n_heads": 16,
        "n_layer": 24,
        "num_layers_trained": 24,
        "act_fn": "gelu",
        "lr_warmup": 0.1,
        "lr": 1e-5,
        "l2_reg": 0.1,
        "epsilon": 1e-8,
        "bert_intermediate_size": 4096,
        "bert_use_pooler": False,
        "max_length": 512,
        "base_model_path": os.path.join("bert", "roberta-model-lg.jl"),
    }
    required_files = [
        {
            "file": os.path.join(
                FINETUNE_BASE_FOLDER, "model", "bert", "roberta-model-lg.jl"
            ),
            "url": urljoin(ROBERTA_BASE_URL, "roberta-model-lg.jl"),
        },
        {
            "file": os.path.join(FINETUNE_BASE_FOLDER, "model", "bert", "dict.txt"),
            "url": urljoin(ROBERTA_BASE_URL, "dict.txt"),
        },
        {
            "file": os.path.join(FINETUNE_BASE_FOLDER, "model", "gpt2", "vocab.bpe"),
            "url": urljoin(GPT2_BASE_URL, "vocab.bpe"),
        },
        {
            "file": os.path.join(FINETUNE_BASE_FOLDER, "model", "gpt2", "encoder.json"),
            "url": urljoin(GPT2_BASE_URL, "encoder.json"),
        },
    ]

   
ZuckerBERT = RoBERTa


class DistilBERT(SourceModel):
    is_bidirectional = True
    encoder = DistilBERTEncoder
    featurizer = bert_featurizer
    settings = {
        "n_embed": 768,
        "n_epochs": 8,
        "n_heads": 12,
        "n_layer": 6,
        "num_layers_trained": 6,
        "act_fn": "gelu",
        "lr_warmup": 0.1,
        "lr": 1e-5,
        "l2_reg": 0.01,
        "bert_use_pooler": False,
        "bert_use_type_embed": False,
        "bert_intermediate_size": 3072,
        "base_model_path": os.path.join("bert", "distillbert.jl"),
    }
    required_files = [
        {
            "file": os.path.join(FINETUNE_BASE_FOLDER, "model", "bert", filename),
            "url": urljoin(BERT_BASE_URL, filename),
        }
        for filename in ["distillbert.jl", "distillbert_vocab.txt"]
    ]


class DistilRoBERTa(SourceModel):
    is_bidirectional = True
    encoder = RoBERTaEncoder
    featurizer = bert_featurizer
    settings = {
        "max_length": 512,
        "n_embed": 768,
        "n_epochs": 8,
        "n_heads": 12,
        "n_layer": 6,
        "num_layers_trained": 6,
        "act_fn": "gelu",
        "lr_warmup": 0.1,
        "lr": 1e-5,
        "l2_reg": 0.01,
        "bert_use_pooler": False,
        "bert_use_type_embed": False,
        "bert_intermediate_size": 3072,
        "base_model_path": os.path.join("bert", "distilroberta.jl"),
    }
    required_files = [
        {
            "file": os.path.join(FINETUNE_BASE_FOLDER, "model", "bert", 'distilroberta.jl'),
            "url": urljoin(BERT_BASE_URL, 'distilroberta.jl'),
        },
        {
            "file": os.path.join(FINETUNE_BASE_FOLDER, "model", "bert", "dict.txt"),
            "url": urljoin(ROBERTA_BASE_URL, "dict.txt"),
        },
        {
            "file": os.path.join(FINETUNE_BASE_FOLDER, "model", "gpt2", "vocab.bpe"),
            "url": urljoin(GPT2_BASE_URL, "vocab.bpe"),
        },
        {
            "file": os.path.join(FINETUNE_BASE_FOLDER, "model", "gpt2", "encoder.json"),
            "url": urljoin(GPT2_BASE_URL, "encoder.json"),
        }
    ]



import logging
import functools

import numpy as np
import tensorflow as tf


from finetune.nn.target_blocks import language_model
from finetune.util.text_generation import sample_with_temperature
from finetune.optimizers.zero_grad import dont_optimize_zeros
from finetune.optimizers.gradient_accumulation import get_grad_accumulation_optimizer
from finetune.optimizers.learning_rate_schedules import schedules
from finetune.optimizers.adamax import AdamaxWOptimizer
from finetune.optimizers.adamw import AdamWOptimizer
from finetune.util.imbalance import class_weight_tensor
from finetune.errors import FinetuneError
from finetune.base_models import GPTModel, GPTModelSmall

LOGGER = logging.getLogger("finetune")

OPTIMIZERS = {"AdamW": AdamWOptimizer, "AdamaxW": AdamaxWOptimizer}


class PredictMode:
    FEATURIZE = "FEAT"
    NORMAL = "NORM"
    PROBAS = "PROBA"
    GENERATE_TEXT = "GEN_TEXT"
    LM_PERPLEXITY = "PERPLEXITY"
    ATTENTION = "ATTENTION"
    SEQUENCE = "SEQUENCE"
    SEQUENCE_PROBAS = "SEQUENCE_PROBA"
    ASSOCIATION = "ASSOCIATION"
    ASSOCIATION_PROBAS = "ASSOCIATION_PROBA"
    EXPLAIN = "EXPLAIN"


def get_model_fn(
        *,
        target_model_fn,
        predict_op,
        predict_proba_op,
        build_target_model,
        build_lm,
        encoder,
        target_dim,
        label_encoder,
        build_explain,
        n_replicas
):
    def language_model_op(X, M, params, featurizer_state):
        language_model_state = language_model(
            X=X,
            M=M,
            config=params,
            embed_weights=featurizer_state["embed_weights"],
            hidden=featurizer_state["sequence_features"],
        )

        lm_logits = language_model_state["logits"]

        lm_logit_mask = np.zeros(
            [1, lm_logits.get_shape().as_list()[-1]], dtype=np.float32
        )
        lm_logit_mask[:, encoder.vocab_size :] = -np.inf

        if "use_extra_toks" in params and not params.use_extra_toks:
            lm_logit_mask[:, encoder.start_token] = -np.inf
            lm_logit_mask[:, encoder.delimiter_token] = -np.inf
            lm_logit_mask[:, encoder.end_token] = -np.inf

        lm_logits += lm_logit_mask
        lm_predict_op = sample_with_temperature(lm_logits, params.lm_temp)
        return lm_predict_op, language_model_state

    def target_model_op(featurizer_state, Y, params, mode, **kwargs):
        weighted_tensor = None
        if params.class_weights is not None:
            weighted_tensor = class_weight_tensor(
                class_weights=params.class_weights,
                target_dim=target_dim,
                label_encoder=label_encoder,
            )
        with tf.variable_scope("model/target"):
            target_model_state = target_model_fn(
                config=params,
                featurizer_state=featurizer_state,
                targets=Y,
                n_outputs=target_dim,
                train=(mode == tf.estimator.ModeKeys.TRAIN),
                max_length=params.max_length,
                class_weights=weighted_tensor,
                **kwargs
            )
        return target_model_state

    def _model_fn(features, labels, mode, params):
        if params.base_model.is_bidirectional and build_lm:
            raise ValueError(
                "Bert models do not support functions that require the language model."
            )
        if not build_target_model:
            lm_loss_coef = 1.0
        else:
            lm_loss_coef = params.lm_loss_coef

        estimator_mode = mode
        train = estimator_mode == tf.estimator.ModeKeys.TRAIN
        X = features["tokens"]
        M = features["mask"]
        task_id = features.get("task_id", None)
        Y = labels
        pred_op = None

        with tf.variable_scope(tf.get_variable_scope()):
            train_loss = 0.0
            featurizer_state = params.base_model.get_featurizer(
                X,
                encoder=encoder,
                config=params,
                train=train,
                explain=build_explain,
            )
            predictions = {
                PredictMode.FEATURIZE: featurizer_state["features"], 
                PredictMode.SEQUENCE: featurizer_state["sequence_features"]
            }
            if params.base_model in [GPTModel, GPTModelSmall]:
                predictions[PredictMode.ATTENTION] = featurizer_state[
                    "attention_weights"
                ]

            if params.base_model in [GPTModel, GPTModelSmall]:
                predictions[PredictMode.ATTENTION] = featurizer_state[
                    "attention_weights"
                ]

            if build_target_model:
                target_model_state = target_model_op(
                    featurizer_state=featurizer_state,
                    Y=Y,
                    params=params,
                    mode=mode,
                    task_id=task_id,
                )
                if (
                    mode == tf.estimator.ModeKeys.TRAIN
                    or mode == tf.estimator.ModeKeys.EVAL
                ) and Y is not None:
                    target_loss = tf.reduce_mean(target_model_state["losses"])
                    train_loss += (1 - lm_loss_coef) * target_loss
                    tf.summary.scalar("TargetModelLoss", target_loss)
                if mode == tf.estimator.ModeKeys.PREDICT or tf.estimator.ModeKeys.EVAL:
                    logits = target_model_state["logits"]
                    predict_params = target_model_state.get("predict_params", {})
                    if "_threshold" in params:
                        predict_params["threshold"] = params._threshold
                    pred_op = predict_op(logits, **predict_params)

                    if type(pred_op) == tuple:
                        pred_op, pred_proba_op = pred_op
                    else:
                        pred_proba_op = predict_proba_op(logits, **predict_params)

                    if type(pred_op) == dict:
                        predictions.update(pred_op)
                        predictions.update(pred_proba_op)
                    else:
                        predictions[PredictMode.NORMAL] = pred_op
                        predictions[PredictMode.PROBAS] = pred_proba_op

                    if build_explain:
                        predictions[PredictMode.EXPLAIN] = target_model_state[
                            "explanation"
                        ]

            if build_lm:
                lm_predict_op, language_model_state = language_model_op(
                    X=X, M=M, params=params, featurizer_state=featurizer_state
                )
                if (
                    mode == tf.estimator.ModeKeys.TRAIN
                    or mode == tf.estimator.ModeKeys.EVAL
                ):
                    lm_loss = tf.reduce_mean(language_model_state["losses"])
                    train_loss += lm_loss_coef * lm_loss
                    tf.summary.scalar("LanguageModelLoss", lm_loss)
                if mode == tf.estimator.ModeKeys.PREDICT:
                    predictions[PredictMode.GENERATE_TEXT] = lm_predict_op
                    predictions[PredictMode.LM_PERPLEXITY] = language_model_state[
                        "perplexity"
                    ]

        if mode == tf.estimator.ModeKeys.TRAIN:
            total_num_steps = params.n_epochs * params.dataset_size // (params.batch_size * n_replicas)
            lr_decay = lambda lr, global_step: tf.maximum(
                0.0,
                lr
                * schedules[params.lr_schedule](
                    tf.to_float(global_step) / total_num_steps, warmup=params.lr_warmup
                ),
            )

            if params.adapter_size is not None:
                norm_variable_scopes = ["b:0", "g:0", "beta:0", "gamma:0"]
                # trained variables include: adapter dense layers, scaling/bias factors, target model, and
                # the bias values in 1dconv if this layer exists (since it also has a 'b' in its name/scope).
                params.trained_variables = [
                    v
                    for v in tf.global_variables()
                    if "adapter" in v.name
                    or "target" in v.name
                    or any(scope in v.name for scope in norm_variable_scopes)
                ]
            else:
                params.trained_variables = [v for v in tf.global_variables()]

            def optimizer(lr):

                Optimizer = OPTIMIZERS.get(params.optimizer, None)
                if Optimizer is None:
                    raise FinetuneError(
                        "Optimizer must be in {}, not {}".format(
                            list(OPTIMIZERS.keys()), params.optimizer
                        )
                    )

                if params.accum_steps > 1:
                    Optimizer = get_grad_accumulation_optimizer(
                        Optimizer, params.accum_steps
                    )

                opt = Optimizer(
                    learning_rate=lr,
                    beta1=params.b1,
                    beta2=params.b2,
                    epsilon=params.epsilon,
                    weight_decay=params.l2_reg * lr,
                )
                decay_var_list = [
                    v
                    for v in tf.global_variables()
                    if len(v.get_shape()) > 1 or params.vector_l2 and "OptimizeLoss" not in v.name
                ]

                if params.adapter_size is not None:
                    decay_var_list = set(params.trained_variables).intersection(
                        decay_var_list
                    )

                opt.apply_gradients = functools.partial(
                    opt.apply_gradients, decay_var_list=decay_var_list
                )

                if params.dont_optimize_zero_gradients:
                    opt = dont_optimize_zeros(opt)

                return opt

            summaries = (
                tf.contrib.layers.OPTIMIZER_SUMMARIES
                if params.summarize_grads
                else None
            )
            train_op = tf.contrib.layers.optimize_loss(
                loss=train_loss,
                global_step=tf.train.get_or_create_global_step(),
                learning_rate=tf.constant(params.lr),
                optimizer=optimizer,
                clip_gradients=float(params.max_grad_norm),
                learning_rate_decay_fn=lr_decay,
                increment_global_step=True,
                summaries=summaries,
                colocate_gradients_with_ops=True,
                variables=params.trained_variables,
            )

        if mode == tf.estimator.ModeKeys.PREDICT:
            return tf.estimator.EstimatorSpec(mode=mode, predictions=predictions)

        if mode == tf.estimator.ModeKeys.TRAIN:
            return tf.estimator.EstimatorSpec(
                mode=mode, loss=train_loss, train_op=train_op
            )

        assert mode == tf.estimator.ModeKeys.EVAL, "The mode is actually {}".format(
            mode
        )
        if params.eval_acc and pred_op is not None:
            LOGGER.info("Adding evaluation metrics, Accuracy")
            labels_dense = tf.argmax(labels, -1)
            metrics = {"Accuracy": tf.metrics.accuracy(tf.argmax(pred_op, -1), labels_dense)}
        else:
            metrics = None

        return tf.estimator.EstimatorSpec(
            mode=mode, loss=train_loss, eval_metric_ops=metrics
        )

    return _model_fn


def get_separate_model_fns(
    target_model_fn,
    predict_op,
    predict_proba_op,
    build_target_model,
    encoder,
    target_dim,
    label_encoder,
    saver,
    portion,
    build_attn,
):
    def _featurizer_model_fn(features, labels, mode, params):
        assert (
            mode == tf.estimator.ModeKeys.PREDICT
        ), "mode MUST be predict - model fns should not be separated on train"
        X = features["tokens"]
        featurizer_state = params.base_model.get_featurizer(
            X,
            encoder=encoder,
            config=params,
            train=False,
        )
        predictions = {
            "features": featurizer_state["features"],
            "sequence_features": featurizer_state["sequence_features"],
            "eos_idx": featurizer_state["eos_idx"],
            "lengths": featurizer_state["lengths"],
        }

        if params.base_model in [GPTModel, GPTModelSmall] and build_attn:
            predictions["attention_weights"] = featurizer_state["attention_weights"]

        return tf.estimator.EstimatorSpec(mode=mode, predictions=predictions)

    if portion == "featurizer":
        return _featurizer_model_fn

    def _target_model_fn(features, labels, mode, params):
        assert (
            mode == tf.estimator.ModeKeys.PREDICT
        ), "Separated estimators are only supported for inference."
        predictions = {}
        featurizer_state = features

        if params.base_model in [GPTModel, GPTModelSmall] and build_attn:
            predictions[PredictMode.ATTENTION] = featurizer_state["attention_weights"]

        with tf.variable_scope("model/target"):
            target_model_state = target_model_fn(
                config=params,
                featurizer_state=featurizer_state,
                targets=None,
                n_outputs=target_dim,
                train=False,
                max_length=params.max_length,
                class_weights=None,
            )

        logits = target_model_state["logits"]
        predict_params = target_model_state.get("predict_params", {})
        if "_threshold" in params:
            predict_params["threshold"] = params._threshold
        pred_op = predict_op(logits, **predict_params)

        if type(pred_op) == tuple:
            pred_op, pred_proba_op = pred_op
        else:
            pred_proba_op = predict_proba_op(logits, **predict_params)

        if type(pred_op) == dict:
            predictions.update(pred_op)
            predictions.update(pred_proba_op)
        else:
            predictions[PredictMode.NORMAL] = pred_op
            predictions[PredictMode.PROBAS] = pred_proba_op

        return tf.estimator.EstimatorSpec(mode=mode, predictions=predictions)

    return _target_model_fn




from abc import ABCMeta, abstractmethod


class SourceModel(metaclass=ABCMeta):
    is_bidirectional = True

    @property
    @abstractmethod
    def encoder(cls):
        raise NotImplementedError

    @property
    @abstractmethod
    def featurizer(cls):
        raise NotImplementedError

    @property
    @abstractmethod
    def settings(cls):
        raise NotImplementedError

    @classmethod
    def get_encoder(cls, **kwargs):
        return cls.encoder(**kwargs)

    @classmethod
    def get_featurizer(cls, X, encoder, config, train=False, reuse=None, **kwargs):
        return cls.featurizer(X, encoder, config, train=train, reuse=reuse, **kwargs)


from finetune.base_models.gpt.model import GPTModel, GPTModelSmall
from finetune.base_models.gpt2.model import GPT2Model, GPT2Model345, GPT2Model762, GPT2Model1558
from finetune.base_models.textcnn.model import TextCNNModel
from finetune.base_models.bert.model import BERTModelCased, BERTModelLargeCased, RoBERTa, RoBERTaLarge, DistilBERT, DistilRoBERTa
from finetune.base_models.tcn.model import TCNModel

# Aliases
GPT = GPTModel
GPT2 = GPT2Small = GPT2Model
GPT2Medium = GPT2Model345
GPT2Large = GPT2Model762
GPT2XL = GPT2Model1558
TextCNN = TextCNNModel
BERT = BERTModelCased
BERTLarge = BERTModelLargeCased
ROBERTA = RoBERTa
ROBERTALarge = RoBERTaLarge
DistilROBERTA = DistilRoBERTa
TCN = TCNModel



import tensorflow as tf
from tensorflow.python.distribute import device_util
from tensorflow.contrib.distribute import ParameterServerStrategy
from tensorflow.contrib.distribute.python.parameter_server_strategy import ParameterServerExtended
from tensorflow.python.distribute import cross_device_ops as cross_device_ops_lib


_LOCAL_CPU = "/device:CPU:0"
_LOCAL_GPU_0 = "/device:GPU:0"




import logging
import os
import os.path
import subprocess
import warnings
from collections import namedtuple
from functools import lru_cache

import numpy as np
from nltk.metrics.distance import edit_distance

import finetune
from finetune.errors import FinetuneError
from finetune.base_models import GPTModel, GPT2Model, BERT, TextCNN

LOGGER = logging.getLogger("finetune")


def finetune_model_path(path):
    return os.path.abspath(
        os.path.join(os.path.dirname(finetune.__file__), "model", path)
    )


def nvidia_device_ids():
    sp = subprocess.Popen(
        ["nvidia-smi", "-L"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    response = sp.communicate()[0]
    gpu_list = response.decode("utf-8").strip().split("\n")
    device_ids = {}
    for i, gpu in enumerate(gpu_list):
        # May be worth logging GPU description
        device_id_str, _, description = gpu.partition(":")
        assert int(device_id_str.split(" ")[-1]) == i
        device_ids[i] = description
    return device_ids


@lru_cache()
def all_gpus(visible_gpus=None):
    """
    Get integer ids of all available GPUs.

    Sample response from nvidia-smi -L:
        GPU 0: GeForce GTX 980 (UUID: GPU-2d683060-957f-d5ad-123c-a5b49b0116d9)
        GPU 1: GeForce GTX 980 (UUID: GPU-7b8496dc-3eaf-8db7-01e7-c4a884f66acf)
        GPU 2: GeForce GTX TITAN X (UUID: GPU-9e01f108-e7de-becd-2589-966dcc1c778f)
    """
    if visible_gpus is not None:
        visible_gpus = [int(gpu) for gpu in visible_gpus]
    try:
        cuda_visible_devices = os.getenv("CUDA_VISIBLE_DEVICES")
        device_ids = nvidia_device_ids()
        mapping = None
        # restricting GPUs based on env vars
        if cuda_visible_devices:
            device_ids = {
                device_id: description
                for device_id, description in device_ids.items()
                if str(device_id) in cuda_visible_devices.split(",")
            }
            mapping = {
                dev_id: i for i, (dev_id, _) in enumerate(sorted(device_ids.items()))
            }

        # restricting GPUs based on config
        if visible_gpus is not None:
            device_ids = {
                device_id: description
                for device_id, description in device_ids.items()
                if device_id in visible_gpus
            }

        LOGGER.info(
            " Visible GPUs: {{{}}}".format(
                ", ".join(
                    [
                        "{}:{}".format(device_id, description.split("(")[0]).strip()
                        for device_id, description in device_ids.items()
                    ]
                )
            )
        )

        if mapping is not None:
            # Resolve these to internal tensorflow device ids.
            # These are equivalent if no visible_devices masking is used
            device_ids = {
                mapping[device_id]: description
                for device_id, description in device_ids.items()
            }

        device_ids = list(device_ids.keys())
    except:
        # Failed to parse out available GPUs properly
        warnings.warn("Failed to find available GPUS.  Falling back to CPU only mode.")
        device_ids = []

    return device_ids


GridSearchable = namedtuple("GridSearchable", "default iterator")


class Settings(dict):
    """
    Model configuration options

    :param base_model: Which base model to use - one of {GPT, GPT2, RoBERTa, BERT, TextCNN, TCN}, imported from finetune.base_models. Defaults to `GPT`.
    :param batch_size: Number of examples per batch, defaults to `2`.
    :param visible_gpus: List of integer GPU ids to spread out computation across, defaults to all available GPUs.
    :param n_epochs: Number of iterations through training data, defaults to `3`.
    :param seed: Random seed to use for repeatability purposes, defaults to `42`.
    :param max_length:  Maximum number of subtokens per sequence. Examples longer than this number will be truncated
        (unless `chunk_long_sequences=True` for SequenceLabeler models). Defaults to `512`.
    :param weight_stddev: Standard deviation of initial weights.  Defaults to `0.02`.
    :param chunk_long_sequences: When True, use a sliding window approach to predict on
        examples that are longer than max length.  The progress bar will display the number of chunks processed rather than the number of examples. Defaults to `True`.  
    :param low_memory_mode: When True, only store partial gradients on forward pass
        and recompute remaining gradients incrementally in order to save memory.  Defaults to `False`.
    :param interpolate_pos_embed: Interpolate positional embeddings when `max_length` differs from it's original value of
        `512`. Defaults to `False`.
    :param embed_p_drop: Embedding dropout probability.  Defaults to `0.1`.
    :param attn_p_drop: Attention dropout probability.  Defaults to `0.1`.
    :param resid_p_drop: Residual layer fully connected network dropout probability.  Defaults to `0.1`.
    :param clf_p_drop: Classifier dropout probability.  Defaults to `0.1`.
    :param l2_reg: L2 regularization coefficient. Defaults to `0.01`.
    :param vector_l2: Whether to apply weight decay regularization to vectors (biases, normalization etc..). Defaults to False.
    :param optimizer: Optimizer to use, current options include AdamW or AdamaxW.
    :param b1: Adam b1 parameter.  Defaults to `0.9`.
    :param b2: Adam b2 parameter.  Defaults to `0.999`.
    :param epsilon: Adam epsilon parameter: Defaults to `1e-8`.
    :param lr_schedule: Learning rate schedule -- see `finetune/optimizers.py` for more options.
    :param lr: Learning rate.  Defaults to `6.25e-5`.
    :param lr_warmup: Learning rate warmup (percentage of all batches to warmup for).  Defaults to `0.002`.
    :param max_grad_norm: Clip gradients larger than this norm. Defaults to `1.0`.
    :param shuffle_buffer_size: How many examples to load into a buffer before shuffling. Defaults to `100`.
    :param dataset_size: Must be specified in order to calculate the learning rate schedule when the inputs provided are generators rather than static datasets.  
    :param accum_steps: Number of updates to accumulate before applying. This is used to simulate a higher batch size.
    :param lm_loss_coef: Language modeling loss coefficient -- a value between `0.0` - `1.0`
        that indicates how to trade off between language modeling loss
        and target model loss.  Usually not beneficial to turn on unless
        dataset size exceeds a few thousand examples.  Defaults to `0.0`.
    :param tsa_schedule: Training Signal Annealing Schedule from 'Unsupervised Data Augmentation for Consistency Training'. One of {"linear_schedule", "exp_schedule", "log_schedule"}.  Defaults to `None`.
    :param summarize_grads: Include gradient summary information in tensorboard.  Defaults to `False`.
    :param val_size: Validation set size if int. Validation set size as percentage of all training data if float.  Defaults to 0.  If value "auto" is provided, validation will not be run by default if n_examples < 50.
        If n_examples > 50, defaults to max(5, min(100, 0.05 * n_examples))
    :param val_interval: Evaluate on validation set after `val_interval` batches.
        Defaults to 4 * val_size / batch_size to ensure that too much time is not spent on validation.
    :param lm_temp: Language model temperature -- a value of `0.0` corresponds to greedy maximum likelihood predictions
        while a value of `1.0` corresponds to random predictions. Defaults to `0.6`.
    :param seq_num_heads: Number of attention heads of final attention layer. Defaults to `16`.
    :param keep_best_model: Whether or not to keep the highest-performing model weights throughout the train. Defaults to `False`.
    :param early_stopping_steps: How many steps to continue with no loss improvement before early stopping. Defaults to `None`.
    :param subtoken_predictions: Return predictions at subtoken granularity or token granularity?  Defaults to `False`.
    :param multi_label_sequences: Use a multi-labeling approach to sequence labeling to allow overlapping labels.
    :param multi_label_threshold: Threshold of sigmoid unit in multi label classifier.
        Can be increased or lowered to trade off precision / recall. Defaults to `0.5`.
    :param autosave_path: Save current best model (as measured by validation loss) to this location. Defaults to `None`.
    :param tensorboard_folder: Directory for tensorboard logs. Tensorboard logs will not be written
        unless tensorboard_folder is explicitly provided. Defaults to `None`.
    :param log_device_placement: Log which device each operation is placed on for debugging purposes.  Defaults to `False`.
    :param allow_soft_placement: Allow tf to allocate an operation to a different device if a device is unavailable.  Defaults to `True`.
    :param save_adam_vars: Save adam parameters when calling `model.save()`.  Defaults to `True`.
    :param num_layers_trained: How many layers to finetune.  Specifying a value less than model's number of layers will train layers starting from model output. Defaults to `12`.
    :param train_embeddings: Should embedding layer be finetuned? Defaults to `True`.
    :param class_weights: One of 'log', 'linear', or 'sqrt'. Auto-scales gradient updates based on class frequency.  Can also be a dictionary that maps from true class name to loss coefficient. Defaults to `None`.
    :param oversample: Should rare classes be oversampled?  Defaults to `False`.
    :param params_device: Which device should gradient updates be aggregated on?
        If you are using a single GPU and have more than 4Gb of GPU memory you should set this to GPU PCI number (0, 1, 2, etc.). Defaults to `"cpu"`.
    :param eval_acc: if True, calculates accuracy and writes it to the tensorboard summary files for valudation runs.
    :param save_dtype: specifies what precision to save model weights with.  Defaults to `np.float32`.
    :param regression_loss: the loss to use for regression models. One of `L1` or `L2`, defaults to `L2`.
    :param prefit_init: if True, fit target model weigths before finetuning the entire model. Defaults to `False`.
    :param debugging_logs: if True, output tensorflow logs and turn off TQDM logging. Defaults to `False`.
    :param val_set: Where it is neccessary to use an explicit validation set, provide it here as a tuple (text, labels)
    :param per_process_gpu_memory_fraction: fraction of the overall amount of memory that each visible GPU should be allocated, defaults to `1.0`.
    :param adapter_size: width of adapter module from 'Parameter Efficient Transfer Learning' paper, if defined. defaults to 'None'.
    """

    def get_grid_searchable(self):
        return self.grid_searchable

    def __init__(self, **kwargs):
        super().__init__()
        self.grid_searchable = {}
        for key, value in kwargs.items():
            self[key] = value

    def __getattr__(self, attr):
        if attr.startswith("__"):
            raise AttributeError

        if attr == "base_model_path":
            full_path = finetune_model_path(self["base_model_path"])
            if os.path.exists(full_path):
                return full_path

        return self[attr]

    def __setitem__(self, key, value):
        if isinstance(value, GridSearchable):
            self.grid_searchable[key] = value.iterator
            value = value.default
        return super().__setitem__(key, value)

    def __setattr__(self, k, v):
        return self.__setitem__(k, v)

    __delattr__ = dict.__delitem__


def did_you_mean(keyword, keyword_pool):
    candidates = list(keyword_pool)
    closest_match_idx = np.argmin(
        [edit_distance(keyword, candidate) for candidate in candidates]
    )
    return candidates[closest_match_idx]


def assert_valid_config(**kwargs):
    expected_keys = set(get_default_config().keys())
    for kwarg in kwargs:
        if kwarg not in expected_keys:
            raise FinetuneError(
                "Unexpected setting configuration: `{}` is an invalid keyword. "
                "Did you mean `{}`?".format(kwarg, did_you_mean(kwarg, expected_keys))
            )


def get_default_config():
    """
    Gets a config object containing all the default parameters for each variant of the model.

    :return: Config object.
    """
    settings = Settings(
        # General Settings
        low_memory_mode=False,
        interpolate_pos_embed=False,
        save_adam_vars=True,
        shuffle_buffer_size=100,
        dataset_size=None,
        batch_size=2,
        visible_gpus=None,  # defaults to all available
        n_epochs=GridSearchable(3, [1, 2, 3, 4]),
        seed=42,
        max_length=512,
        weight_stddev=0.02,
        save_dtype=None,
        val_set=None,
        per_process_gpu_memory_fraction=None,
        adapter_size=None,  # from Parameter Efficient Transfer Learning paper
        
        # Regularization
        embed_p_drop=0.1,
        attn_p_drop=0.1,
        resid_p_drop=0.1,
        clf_p_drop=0.1,
        l2_reg=GridSearchable(0.01, [0.0, 0.1, 0.01, 0.001]),
        vector_l2=False,
        
        # Early Stopping and Validation
        autosave_path=None,
        keep_best_model=False,
        early_stopping_steps=None,
        min_secs_between_eval=60,
        eval_acc=False,
        val_size=0.,
        val_interval=None,

        # Debugging
        log_device_placement=False,
        soft_device_placement=True,
        tensorboard_folder=None,
        summarize_grads=False,
        debugging_logs=False,

        # Partial Fitting
        num_layers_trained=12,
        train_embeddings=True,

        # Class Imbalance
        class_weights=None,
        oversample=False,
        params_device="cpu",

        # Optimization Params
        optimizer="AdamW",
        b1=0.9,
        b2=0.999,
        epsilon=1e-8,
        lr_schedule="warmup_linear",
        lr=GridSearchable(6.25e-5, [6.25e-4, 6.25e-5, 6.25e-6]),
        lr_warmup=0.002,
        max_grad_norm=1.0,
        prefit_init=False,
        accum_steps=1,
        tsa_schedule=None,

        # MTL
        tasks=None,
        dont_optimize_zero_gradients=False,

        # Language Model Settings
        lm_loss_coef=0.0,
        lm_temp=0.6,

        # Sequence Labeling
        seq_num_heads=16,
        pad_token="<PAD>",
        pad_idx=None,
        subtoken_predictions=False,
        multi_label_sequences=False,
        multi_label_threshold=0.5,
        chunk_long_sequences=True,
        add_eos_bos_to_chunk=True,
        filter_empty_examples=False, 

        # Regression Params
        regression_loss="L2",

        # Association Params
        viable_edges=None,
        association_types=None,
        assocation_loss_weight=100.0,

        # Location of model weights
        base_model=GPTModel,
        base_model_path=None,

        # Possible `SourceModel` specific settings
        n_heads=None,
        n_layer=None,
        act_fn=None,
        n_embed=None,

        # for TCN SourceModel only
        n_filter=None,
        kernel_size=None,

        # for TextCNN SourceModel only
        kernel_sizes=None,
        num_filters_per_size=None,
        n_embed_featurizer=None,  # needed because the dimensions CNN output are different from the embedding dimensions

        # BERT only
        bert_intermediate_size=None,
        bert_use_pooler=True,
        bert_use_type_embed=True,

        # Auxiliary Information
        use_auxiliary_info=False,
        n_context_embed=32,
        default_context=None,
        context_dim=None
    )
    return settings


def get_config(error_on_invalid_keywords=True, **kwargs):
    """
    Gets a config object containing all the default parameters for each variant of the model.

    :param **kwargs: Keyword arguments to override default values.
    :return: Config object.    """
    if error_on_invalid_keywords:
        assert_valid_config(**kwargs)
    config = get_default_config()
    config.base_model = kwargs.get("base_model", config.base_model)
    config.update(config.base_model.settings)
    config.update(kwargs)
    return config


def cpu_config(**kwargs):
    config = get_config(**kwargs)
    config.visible_gpus = []
    config.update(kwargs)
    return config



import tensorflow as tf
from finetune.errors import FinetuneError

# From Unsupervised Data Augmentation for Consistency Training, Xie et al. 2019
def get_tsa_threshold(schedule, global_step, num_train_steps, start, end):
    if schedule not in ["linear_schedule", "exp_schedule", "log_schedule"]:
        raise FinetuneError(
            'Invalid TSA schedule set in config. Must be one of ["linear_schedule", "exp_schedule", "log_schedule"].'
        )
    training_progress = tf.to_float(global_step) / tf.to_float(num_train_steps)
    if schedule == "linear_schedule":
        threshold = training_progress
    elif schedule == "exp_schedule":
        scale = 5
        threshold = tf.exp((training_progress - 1) * scale)
        # [exp(-5), exp(0)] = [1e-2, 1]
    elif schedule == "log_schedule":
        scale = 5
        # [1 - exp(0), 1 - exp(-5)] = [0, 0.99]
        threshold = 1 - tf.exp((-training_progress) * scale)
    return threshold * (end - start) + start


def tsa_loss(n_targets, config, clf_losses, clf_logits, targets):
    with tf.variable_scope("tsa"):
        start = 1 / n_targets
        global_step = tf.train.get_or_create_global_step()
        total_num_steps = config.n_epochs * config.dataset_size // config.batch_size
        tsa_threshold = get_tsa_threshold(
            config.tsa_schedule, global_step, total_num_steps, start, 1
        )

        clf_logits = tf.nn.log_softmax(clf_logits)
        multiplied = targets * tf.exp(clf_logits)
        correct_label_probs = tf.reduce_sum(multiplied, axis=1)
        larger_than_threshold = tf.greater(correct_label_probs, tsa_threshold)
        loss_mask = tf.ones_like(clf_losses, dtype=clf_losses.dtype)

        loss_mask = loss_mask * (1 - tf.cast(larger_than_threshold, tf.float32))
        loss_mask = tf.stop_gradient(loss_mask)
        clf_losses = tf.multiply(clf_losses, loss_mask)

        return clf_logits, clf_losses



import tensorflow as tf
import numpy as np
from scipy.stats import mode

from finetune.base import BaseModel
from finetune.encoding.target_encoders import OrdinalRegressionEncoder
from finetune.nn.target_blocks import ordinal_regressor
from finetune.input_pipeline import BasePipeline
from finetune.target_models.comparison_regressor import ComparisonRegressionPipeline
from finetune.target_models.comparison import Comparison


class OrdinalRegressionPipeline(BasePipeline):
    def _target_encoder(self):
        return OrdinalRegressionEncoder()


class ComparisonOrdinalRegressionPipeline(ComparisonRegressionPipeline):
    def _target_encoder(self):
        return OrdinalRegressionEncoder()


class OrdinalRegressor(BaseModel):
    """
    Classifies a document into two or more ordered categories.

    For a full list of configuration options, see `finetune.config`.

    :param config: A config object generated by `finetune.config.get_config` or None (for default config).
    :param \**kwargs: key-value pairs of config items to override.
    """

    def __init__(self, shared_threshold_weights=True, **kwargs):
        super().__init__(**kwargs)
        self.config.shared_threshold_weights = shared_threshold_weights

    def _get_input_pipeline(self):
        return OrdinalRegressionPipeline(self.config)

    def featurize(self, X):
        """
        Embeds inputs in learned feature space. Can be called before or after calling :meth:`finetune`.

        :param X: list or array of text to embed.
        :returns: np.array of features of shape (n_examples, embedding_size).
        """
        return self._featurize(X)

    def predict(self, X):
        """
        Produces a list of most likely class labels as determined by the fine-tuned model.

        :param X: list or array of text to embed.
        :returns: list of class labels.
        """
        all_labels = []
        for _, start_of_doc, end_of_doc, label, _ in self.process_long_sequence(X):
            if start_of_doc:
                # if this is the first chunk in a document, start accumulating from scratch
                doc_labels = []

            doc_labels.append(label)

            if end_of_doc:
                # last chunk in a document
                # since we are working with discrete boundaries, and have no probabilities to average across, we take the mode                
                doc_labels = np.asarray(doc_labels)
                doc_mode = mode(doc_labels).mode
                label = self.input_pipeline.label_encoder.inverse_transform(np.asarray([doc_mode]))
                all_labels.append(label.tolist())
        return all_labels

    def predict_proba(self, X):
        """
        Produces a probability distribution over classes for each example in X.

        :param X: list or array of text to embed.
        :returns: list of dictionaries.  Each dictionary maps from a class label to its assigned class probability.
        """
        raise AttributeError("`Regressor` model does not support `predict_proba`.")

    def finetune(self, X, Y=None, batch_size=None):
        """
        :param X: list or array of text.
        :param Y: floating point targets
        :param batch_size: integer number of examples per batch. When N_GPUS > 1, this number
                           corresponds to the number of training examples provided to each GPU.
        """
        return super().finetune(X, Y=Y, batch_size=batch_size)

    @staticmethod
    def _target_model(
        config, featurizer_state, targets, n_outputs, train=False, reuse=None, **kwargs
    ):
        return ordinal_regressor(
            hidden=featurizer_state["features"],
            targets=targets,
            n_targets=n_outputs,
            config=config,
            train=train,
            reuse=reuse,
            shared_threshold_weights=config.shared_threshold_weights,
            **kwargs
        )

    def _predict_op(self, logits, **kwargs):
        return logits

    def _predict_proba_op(self, logits, **kwargs):
        return logits


class ComparisonOrdinalRegressor(OrdinalRegressor):
    """
    Compares two documents and classifies into two or more ordered categories.

    For a full list of configuration options, see `finetune.config`.

    :param config: A config object generated by `finetune.config.get_config` or None (for default config).
    :param \**kwargs: key-value pairs of config items to override.
    """
    defaults = {"chunk_long_sequences": False}

    def predict(self, pairs):
        """
        Produces a floating point prediction determined by the fine-tuned model.


        :param pairs: Array of text, shape [batch, 2]
        :returns: list of floats, shape [batch]
        """
        return list(Comparison.predict(self, pairs))


    def _get_input_pipeline(self):
        return ComparisonOrdinalRegressionPipeline(self.config)

    @staticmethod
    def _target_model(
        config, featurizer_state, targets, n_outputs, train=False, reuse=None, **kwargs
    ):
        featurizer_state["sequence_features"] = tf.abs(
            tf.reduce_sum(featurizer_state["sequence_features"], 1)
        )
        featurizer_state["features"] = tf.abs(
            tf.reduce_sum(featurizer_state["features"], 1)
        )
        return ordinal_regressor(
            hidden=featurizer_state["features"],
            targets=targets,
            n_targets=n_outputs,
            config=config,
            shared_threshold_weights=config.shared_threshold_weights,
            train=train,
            reuse=reuse,
            **kwargs
        )




import unittest
import os.path

import numpy as np
from tensorflow.data import Dataset

from finetune.encoding.input_encoder import EncodedOutput
from finetune.model import PredictMode
from finetune.base_models import GPT, GPT2, BERT
from finetune import Classifier, MultiFieldClassifier

DIRECTORY = os.path.abspath(os.path.dirname(__file__))


class TestActivationParity(unittest.TestCase):

    MULTIFIELD_TEST_DATA = [[
        "Rick grew up in a troubled household. He never found good support in family, "
        "and turned to gangs. It wasn't long before Rick got shot in a robbery. The "
        "incident caused him to turn a new leaf.", 
        "He is happy now."
    ]]
    TEST_DATA = ["this is a test"]

    def test_gpt_multifield_parity(self):
        model = MultiFieldClassifier()
        np.testing.assert_allclose(
            model.featurize(self.MULTIFIELD_TEST_DATA)[0], 
            np.load(
                os.path.join(
                    DIRECTORY, 
                    'data/test-gpt-multifield-activations.npy'
                )
            ),
            atol=1e-1
        )

    def test_gpt2_featurize(self):
        model = Classifier(base_model=GPT2)
        
        def dataset_encoded():
            yield {"tokens": arr_encoded.token_ids, "mask": arr_encoded.mask}

        def get_input_fn():
            types, shapes = model.input_pipeline.feed_shape_type_def()
            tf_dataset = Dataset.from_generator(dataset_encoded, types[0], shapes[0])
            return tf_dataset.batch(1)

        encoded = model.input_pipeline.text_encoder._encode(self.TEST_DATA)
        encoded = EncodedOutput(token_ids=encoded.token_ids[0])
        estimator, hooks = model.get_estimator(force_build_lm=False)
        predict = estimator.predict(
            input_fn=get_input_fn, predict_keys=[PredictMode.SEQUENCE], hooks=hooks
        )
        arr_encoded = model.input_pipeline._array_format(encoded)
        sequence_features = next(predict)[PredictMode.SEQUENCE]

        np.testing.assert_allclose(
            sequence_features[:len(encoded.token_ids),:],
            np.load(
                os.path.join(
                    DIRECTORY, 
                    'data/test-gpt2-activations.npy'
                )
            ),
            atol=1e-1
        )

    def test_bert_featurize(self):
        model = Classifier(base_model=BERT)
        np.testing.assert_allclose(
            model.featurize(self.TEST_DATA)[0], 
            np.load(
                os.path.join(
                    DIRECTORY, 
                    'data/test-bert-activations.npy'
                )
            ),
            atol=1e-1
        )

    


import os
import urllib
import urllib.request
from pathlib import Path

import finetune

FINETUNE_BASE_FOLDER = os.path.dirname(finetune.__file__)
GPT_BASE_URL     = "https://s3.amazonaws.com/bendropbox/gpt/"
GPT2_BASE_URL    = "https://s3.amazonaws.com/bendropbox/gpt2/"
BERT_BASE_URL    = "https://s3.amazonaws.com/bendropbox/bert/"
ROBERTA_BASE_URL = "https://s3.amazonaws.com/bendropbox/roberta/"

def download_data_if_required(base_model):
    """ Pulls the pre-trained model weights from Github if required. """

    for file_obj in base_model.required_files:
        path = Path(file_obj['file'])
        if not path.exists():
            path.parent.mkdir(parents=True, exist_ok=True)
            print("Downloading: {}".format(path.name))
            data = urllib.request.urlopen(file_obj['url']).read()
            with path.open('wb') as f:
                f.write(data)



import os
import unittest
import logging
from copy import copy
from pathlib import Path
import codecs
import json

import tensorflow as tf
# required for tensorflow logging control
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

from finetune import MultipleChoice


class TestQuestionAnswer(unittest.TestCase):

    def test_reasonable_predictions(self):
        model = MultipleChoice(n_epochs=100, val_size=0, max_length=64, batch_size=3, lr=2e-5, lr_schedule="none",
                               val_interval=10000, embed_p_drop=0., attn_p_drop=0., resid_p_drop=0, clf_p_drop=0)
        questions = [
            "Dog, cat, fish, orange, what is the odd one out?",
            "Stocks, Futures, Money, Chicken, what is the odd one out?",
            "England, US, Finland, Penguin, what is the odd one out?",
            "Orange, Yellow, Tiger, Purple, what is the odd one out?",
            "Computer, Cellphone, Coffee, Telephone, what is the odd one out?",
            "Boat, car, chicken, train, what is the odd one out?",
            "Walk, run, william, jump, what is the odd one out?"
        ]

        answers = [
            ["orange", "Dog", "fish", "cat"],
            ["Chicken", "Stocks", "Futures", "Money"],
            ["Penguin", "England", "US", "Finland"],
            ["Tiger", "Orange", "Yellow", "Purple"],
            ["Coffee", "Computer", "Cellphone", "Telephone"],
            ["train", "Boat", "car", "train"],
            ["william", "Walk", "run", "jump"],
        ]

        model.finetune(questions, answers, ["orange", "Chicken", "Penguin", "Tiger", "Coffee", "train", "william"])

        self.assertEqual(["orange"], model.predict(["Dog, cat, fish, orange, what is the odd one out?"],
                                                   [["orange", "Dog", "fish", "cat"]]))



import os
from urllib.parse import urljoin

from finetune.base_models import SourceModel
from finetune.base_models.gpt.encoder import GPTEncoder
from finetune.base_models.gpt2.encoder import GPT2Encoder
from finetune.base_models.textcnn.featurizer import textcnn_featurizer
from finetune.util.download import GPT2_BASE_URL, FINETUNE_BASE_FOLDER


class TextCNNModel(SourceModel):
    is_bidirectional = False
    encoder = GPT2Encoder
    featurizer = textcnn_featurizer
    kernel_sizes = [2, 3, 4]
    settings = {
        "n_embed_featurizer": 768,
        "n_layer": 1,
        "n_epochs": 100,
        "keep_best_model": True,
        "early_stopping_steps": 100,
        "val_size": "auto",
        "num_layers_trained": 1,
        "kernel_sizes": kernel_sizes,
        "num_filters_per_size": 2,
        "n_embed": len(kernel_sizes) * 2,
        "act_fn": "gelu",
        "train_embeddings": True,
        "lr": 0.01,
        "seq_num_heads": len(kernel_sizes) * 2,
        "base_model_path": os.path.join("gpt2", "model-sm.jl"),
    }
    required_files = [
        {
            "file": os.path.join(FINETUNE_BASE_FOLDER, "model", "gpt2", filename),
            "url": urljoin(GPT2_BASE_URL, filename),
        }
        for filename in ["encoder.json", "vocab.bpe", "model-sm.jl"]
    ]



from tensorflow.python.ops import resource_variable_ops
from tensorflow.python.framework import ops
from tensorflow.contrib.opt.python.training.weight_decay_optimizers import AdamWOptimizer as AdamWOptimizerBroken

import tensorflow as tf


class AdamWOptimizer(AdamWOptimizerBroken):
    def _resource_scatter_add(self, x, i, v, _=None):
        with ops.control_dependencies(
                [resource_variable_ops.resource_scatter_add(x.handle, i, v)]):
            return tf.convert_to_tensor(x)



import os
import unittest
import warnings

# prevent excessive warning logs
warnings.filterwarnings('ignore')
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

from sklearn.model_selection import train_test_split

from finetune import MultiTask, Classifier, Comparison
from finetune.datasets.stanford_sentiment_treebank import StanfordSentimentTreebank
from finetune.datasets.quora_similarity import QuoraDuplicate
from finetune.config import finetune_model_path
SST_FILENAME = "SST-binary.csv"


class TestMTL(unittest.TestCase):

    def test_fit_predict(self):
        dataset = StanfordSentimentTreebank(nrows=50).dataframe
        q_dataset = QuoraDuplicate(nrows=50).dataframe

        model = MultiTask(
            tasks={
                "sst": Classifier,
                "qqp": Comparison
            },
            n_epochs=2,
            optimizer="AdamaxW",
            max_length=200,
        )

        q_X1, q_X2, q_Y = q_dataset.Text1.values, q_dataset.Text2.values, q_dataset.Target.values

        trainX, testX, trainY, testY = train_test_split(
            dataset.Text.values, dataset.Target.values, test_size=0.3, random_state=42
        )

        model.fit(
            {
                "sst": trainX,
                "qqp": list(zip(q_X1, q_X2)),
            },
            {
                "sst": trainY,
                "qqp": q_Y,
            }
        )

        model.featurize(
            {
                "sst": testX,
                "qqp": list(zip(q_X1, q_X2))[:10],
            }
        )

        preds = model.predict(
            {
                "sst": testX,
                "qqp": list(zip(q_X1, q_X2))[:10],
            }
        )
        self.assertIn("sst", preds)
        self.assertIn("qqp", preds)

        model.create_base_model("./test_base_mtl.jl", exists_ok=True)
        model = Classifier(base_model_path="./test_base_mtl.jl", max_length=200)
        model.fit(trainX, trainY)

        os.remove(finetune_model_path("./test_base_mtl.jl"))



import functools

import numpy as np
import tensorflow as tf

from finetune.optimizers.recompute_grads import recompute_grad
from finetune.util.shapes import shape_list, lengths_from_eos_idx
from finetune.nn.activations import act_fns
from finetune.nn.nn_utils import dropout, norm


def mask_attn_weights(w):
    n = shape_list(w)[-1]
    b = tf.matrix_band_part(tf.ones([n, n]), -1, 0)
    b = tf.reshape(b, [1, 1, n, n])
    w = w * b + -1e9 * (1 - b)
    return w


def explain_mask_attn_weights(w):
    # w is [batch, heads, n, n]
    # lengths is [batch]
    batch, _, _, n = shape_list(w)
    seq = n // 2
    main_mask = tf.matrix_band_part(tf.ones([seq, seq]), -1, 0)
    top = tf.expand_dims(
        tf.concat((main_mask, tf.zeros([seq, seq])), 1), 0
    )  # 1, seq, 2 * seq
    clf_to_clf_mask = tf.eye(seq)
    bottom = tf.expand_dims(
        tf.concat((main_mask, clf_to_clf_mask), 1), 0
    )  # 1, seq, 2 * seq
    m = tf.concat((top, bottom), 1)
    b = tf.reshape(m, [1, 1, n, n])
    w = w * b + -1e9 * (1 - b)
    return w


def attn_weights(q, k, v, scale=False, mask=True, explain=False):
    w = tf.matmul(q, k)

    if scale:
        n_state = shape_list(v)[-1]
        w = w * tf.rsqrt(tf.cast(n_state, tf.float32))

    if mask:
        if explain:
            w = explain_mask_attn_weights(w)
        else:
            w = mask_attn_weights(w)
    w = tf.nn.softmax(w)
    return w


def split_states(x, n):
    x_shape = shape_list(x)
    m = x_shape[-1]
    new_x_shape = x_shape[:-1] + [n, m // n]
    return tf.reshape(x, new_x_shape)


def merge_states(x):
    x_shape = shape_list(x)
    new_x_shape = x_shape[:-2] + [np.prod(x_shape[-2:])]
    return tf.reshape(x, new_x_shape)


def split_heads(x, n, k=False):
    if k:
        return tf.transpose(split_states(x, n), [0, 2, 3, 1])
    else:
        return tf.transpose(split_states(x, n), [0, 2, 1, 3])


def merge_heads(x):
    return merge_states(tf.transpose(x, [0, 2, 1, 3]))


def conv1d(
    x,
    scope,
    nf,
    rf,
    w_init=tf.random_normal_initializer(stddev=0.02),
    b_init=tf.constant_initializer(0),
    pad="VALID",
    train=False,
):
    with tf.variable_scope(scope):
        nx = shape_list(x)[-1]
        w = tf.get_variable("w", [rf, nx, nf], initializer=w_init)
        b = tf.get_variable("b", [nf], initializer=b_init)
        if rf == 1:  # faster 1x1 conv
            c = tf.reshape(
                tf.matmul(tf.reshape(x, [-1, nx]), tf.reshape(w, [-1, nf])) + b,
                shape_list(x)[:-1] + [nf],
            )
        else:  # was used to train LM
            c = tf.nn.conv1d(x, w, stride=1, padding=pad) + b
        return c


def multihead_qkv(x, n_state, n_head, train, explain=False):
    c = conv1d(x, "c_attn", n_state * 3, 1, train=train)
    q, k, v = tf.split(c, 3, 2)
    q = split_heads(q, n_head)
    k = split_heads(k, n_head, k=True)
    v = split_heads(v, n_head)
    return q, k, v


def attn(
    x,
    scope,
    n_state,
    n_head,
    resid_pdrop,
    attn_pdrop,
    train=False,
    scale=False,
    mask=True,
    explain=False,
):
    assert n_state % n_head == 0
    with tf.variable_scope(scope):
        q, k, v = multihead_qkv(x, n_state, n_head, train, explain)

        w = attn_weights(q, k, v, scale=scale, mask=mask, explain=explain)
        w = dropout(w, attn_pdrop, train)

        a = tf.matmul(w, v)

        a = merge_heads(a)
        a = conv1d(a, "c_proj", n_state, 1, train=train)
        a = dropout(a, resid_pdrop, train)
        return a


def mlp(x, scope, n_state, act_fn, resid_pdrop, train=False):
    with tf.variable_scope(scope):
        nx = shape_list(x)[-1]
        act = act_fns[act_fn]
        h = act(conv1d(x, "c_fc", n_state, 1, train=train))
        h2 = conv1d(h, "c_proj", nx, 1, train=train)
        h2 = dropout(h2, resid_pdrop, train)
        return h2


def create_initializer(initializer_range=0.001):
    """Creates a `truncated_normal_initializer` with the given range."""
    return tf.truncated_normal_initializer(stddev=initializer_range)


def adapter(X, adapter_size, nx, train=False, hidden_dropout_prob=0.1):
    down_projection = tf.layers.dense(
        X,
        adapter_size,
        activation="sigmoid",
        kernel_initializer=create_initializer(0.001),
    )
    down_projection = dropout(down_projection, hidden_dropout_prob, train)
    up_projection = tf.layers.dense(
        down_projection, nx, kernel_initializer=create_initializer(0.001)
    )
    return up_projection + X


def block(
    x,
    n_head,
    act_fn,
    adptr_size,
    resid_pdrop,
    attn_pdrop,
    scope,
    train=False,
    scale=False,
    explain=False,
):
    with tf.variable_scope(scope):
        nx = shape_list(x)[-1]
        a = attn(
            x,
            "attn",
            nx,
            n_head,
            resid_pdrop,
            attn_pdrop,
            train=train,
            scale=scale,
            explain=explain,
        )
        if adptr_size is not None:
            with tf.variable_scope("attn_adapter"):
                a = adapter(a, adptr_size, nx, train)
        n = norm(x + a, "ln_1")
        m = mlp(n, "mlp", nx * 4, act_fn, resid_pdrop, train=train)
        if adptr_size is not None:
            with tf.variable_scope("dense_adapter"):
                m = adapter(m, adptr_size, nx, train)
        h = norm(n + m, "ln_2")
        return h


def embed(X, we):
    e = tf.gather(we, X)
    h = tf.reduce_sum(e, 2)
    return h


def add_explain_tokens(X, max_length, pool_idx):
    flat_x = tf.reshape(X[:, :, :1], [-1, 1])
    flat_pos = tf.minimum(
        X[:, :, 1:] + 1, max_length - 1
    )  # + 1 to offset for start token
    clf_tok = tf.gather(
        flat_x, tf.range(shape_list(X)[0], dtype=tf.int32) * max_length + pool_idx
    )
    clf_tok_x_seq = tf.tile(tf.expand_dims(clf_tok, 1), [1, max_length, 1])
    clf_tok_x_seq_w_pos = tf.concat((clf_tok_x_seq, flat_pos), -1)
    return tf.concat((X, clf_tok_x_seq_w_pos), 1)


def gpt_featurizer(
    X,
    encoder,
    config,
    train=False,
    reuse=None,
    explain=False,
    **kwargs
):
    """
    The transformer element of the finetuning model. Maps from tokens ids to a dense, embedding of the sequence.

    :param X: A tensor of token indexes with shape [batch_size, sequence_length, token_idx]
    :param encoder: A TextEncoder object.
    :param config: A config object, containing all parameters for the featurizer.
    :param train: If this flag is true, dropout and losses are added to the graph.
    :param reuse: Should reuse be set within this scope.
    :return: A dict containing;
        embed_weights: the word embedding matrix.
        features: The output of the featurizer_final state.
        sequence_features: The output of the featurizer at each timestep.
    """
    initial_shape = tf.shape(X)
    X = tf.reshape(X, shape=tf.concat(([-1], initial_shape[-2:]), 0))

    with tf.variable_scope("model/featurizer", reuse=reuse):
        embed_weights = tf.get_variable(
            name="we",
            shape=[encoder.vocab_size + config.max_length, config.n_embed],
            initializer=tf.random_normal_initializer(stddev=config.weight_stddev),
        )
        if config.train_embeddings:
            embed_weights = dropout(embed_weights, config.embed_p_drop, train)
        else:
            embed_weights = tf.stop_gradient(embed_weights)

        X = tf.reshape(X, [-1, config.max_length, 2])

        clf_token = encoder.end_token
        pool_idx = tf.cast(tf.argmax(tf.cast(tf.equal(X[:, :, 0], clf_token), tf.float32), 1), tf.int32)

        if explain:
            X = add_explain_tokens(X, config.max_length, pool_idx)

        h = embed(X, embed_weights)
        for layer in range(config.n_layer):
            if (
                (config.n_layer - layer) == config.num_layers_trained
                and config.num_layers_trained != config.n_layer
                and config.adapter_size is None
            ):
                h = tf.stop_gradient(h)
                train_layer = False
            else:
                train_layer = train

            with tf.variable_scope("h%d_" % layer):
                block_fn = functools.partial(
                    block,
                    n_head=config.n_heads,
                    act_fn=config.act_fn,
                    resid_pdrop=config.resid_p_drop,
                    attn_pdrop=config.attn_p_drop,
                    scope="h%d" % layer,
                    train=train_layer,
                    scale=True,
                    explain=explain,
                    adptr_size=config.adapter_size,
                )
                if config.low_memory_mode and train_layer:
                    block_fn = recompute_grad(block_fn, use_entire_scope=True)
                if layer < config.n_layer - 1:
                    h = block_fn(h)
                else:
                    h_out = block_fn(h)

            # get the attention weights from the last layer
            if layer == config.n_layer - 1:
                with tf.variable_scope("h%d_/h%d/attn" % (layer, layer), reuse=True):
                    q, k, v = multihead_qkv(
                        h, n_state=shape_list(h)[-1], n_head=config.n_heads, train=train
                    )
                    w = attn_weights(q, k, v, scale=True)

        if explain:
            explain_out = h_out[:, initial_shape[1] :]
            explain_out = tf.reshape(
                explain_out, shape=tf.concat((initial_shape[:-1], [config.n_embed]), 0)
            )
            h_out = h_out[:, : initial_shape[1]]

        # Use hidden state at classifier token as input to final proj. + softmax
        clf_h = tf.reshape(h_out, [-1, config.n_embed])  # [batch * seq_len, embed]
        clf_h = tf.gather(
            clf_h,
            tf.range(shape_list(X)[0], dtype=tf.int32) * config.max_length + pool_idx,
        )
        clf_h = tf.reshape(
            clf_h, shape=tf.concat((initial_shape[:-2], [config.n_embed]), 0)
        )
        seq_feats = tf.reshape(
            h, shape=tf.concat((initial_shape[:-1], [config.n_embed]), 0)
        )

        lengths = lengths_from_eos_idx(eos_idx=pool_idx, max_length=shape_list(X)[0])

        out = {
            "embed_weights": embed_weights,
            "features": clf_h,
            "sequence_features": seq_feats,
            "eos_idx": pool_idx,
            "lengths": lengths,
            "attention_weights": w,  # [n_heads, seq_len, seq_len]
        }

        if explain:
            out["explain_out"] = explain_out
        return out



import os
import unittest
import warnings

# prevent excessive warning logs
warnings.filterwarnings('ignore')
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import numpy as np
import random

from sklearn.model_selection import train_test_split
from finetune import ComparisonRegressor


class TestComparisonRegression(unittest.TestCase):
    n_sample = 100

    def default_config(self, **kwargs):
        d = dict(
            batch_size=2,
            max_length=16,
            n_epochs=3,
            val_size=0.,
            l2_reg=0.,
            interpolate_pos_embed=False,
            low_memory_mode=True,
        )
        d.update(kwargs)
        return d

    def setUp(self):
        random.seed(42)
        np.random.seed(42)

    def test_reasonable_predictions(self):
        """
        Ensure model training does not error out
        Ensure model returns predictions of the right type
        Test model loss at least outperforms naive baseline
        """
        model = ComparisonRegressor(**self.default_config())

        # fake dataset generation
        animals = ["dog", "cat", "horse", "cow", "pig", "sheep", "goat", "chicken", "guinea pig", "donkey", "turkey", "duck", "camel", "goose", "llama", "rabbit", "fox"]
        numbers = ["one", "two", "three", "four", "five", "six", "seven", "eight", "nine", "ten", "eleven", "twelve", "thirteen", "fourteen", "fifteen", "sixteen"]

        n_per = 150
        similar = []
        different = []
        for dataset in [animals, numbers]:
            for i in range(n_per // 2):
                similar.append([random.choice(dataset), random.choice(dataset)])
        for i in range(n_per):
            different.append([random.choice(animals), random.choice(numbers)])

        targets = np.asarray([1] * len(similar) + [0] * len(different))
        data = similar + different

        x_tr, x_te, t_tr, t_te = train_test_split(data, targets, test_size=0.3, random_state=42)
        model.finetune(x_tr, t_tr)

        predictions = model.predict(x_te)
        mse = np.mean([(pred - true)**2 for pred, true in zip(predictions, t_te)])
        naive_baseline = max(np.mean(targets == 1), np.mean(targets == 0))
        naive_baseline_mse = np.mean([(naive_baseline - true)**2 for true in t_te])
        self.assertIsInstance(predictions, np.ndarray)
        self.assertIsInstance(predictions[0], np.float32)
        self.assertGreater(naive_baseline_mse, mse)


if __name__ == '__main__':
    unittest.main()



import os
import unittest
import warnings
from pathlib import Path

# prevent excessive warning logs 
warnings.filterwarnings('ignore')
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import tensorflow as tf
import pandas as pd
import numpy as np

from finetune import MultiFieldClassifier, MultiFieldRegressor, Regressor
from finetune.config import get_config
from finetune.datasets import generic_download

SST_FILENAME = "SST-binary.csv"


class TestModel(unittest.TestCase):
    n_sample = 100
    dataset_path = os.path.join(
        'Data', 'Classify', SST_FILENAME
    )

    @classmethod
    def _download_sst(cls):
        """
        Download Stanford Sentiment Treebank to data directory
        """
        path = Path(cls.dataset_path)
        if path.exists():
            return

        path.mkdir(parents=True, exist_ok=True)
        generic_download(
            url="https://s3.amazonaws.com/enso-data/SST-binary.csv",
            text_column="Text",
            target_column="Target",
            filename=SST_FILENAME
        )

    @classmethod
    def setUpClass(cls):
        cls._download_sst()

    def setUp(self):
        try:
            os.mkdir("tests/saved-models")
        except FileExistsError:
            warnings.warn("tests/saved-models still exists, it is possible that some test is not cleaning up properly.")
            pass
        self.save_file = 'tests/saved-models/test-save-load.jl'
        self.dataset = pd.read_csv(self.dataset_path)
        train_sample = self.dataset.sample(n=self.n_sample)
        valid_sample = self.dataset.sample(n=self.n_sample)
        self.text_data_train = [[x] * 3 for x in train_sample.Text.values.tolist()]
        self.text_data_valid = [[x] * 3 for x in valid_sample.Text.values.tolist()]
        self.train_targets = train_sample.Target

    def test_multifield_classify(self):
        """
        Ensure fit predict works on classification with multi inputs
        Ensure saving + loading does not cause errors
        Ensure saving + loading does not change predictions
        """
        self.model = MultiFieldClassifier()
        self.model.fit(self.text_data_train, self.train_targets)
        predictions = self.model.predict(self.text_data_valid)
        self.model.save(self.save_file)
        model = MultiFieldClassifier.load(self.save_file)
        new_predictions = model.predict(self.text_data_valid)
        for new_pred, old_pred in zip(new_predictions, predictions):
            self.assertEqual(new_pred, old_pred)
    
    def test_cached_predict(self):
        """
        Ensure fit predict works on classification with multi inputs
        Ensure saving + loading does not cause errors
        Ensure saving + loading does not change predictions
        """
        self.model = MultiFieldClassifier()
        self.model.fit(self.text_data_train, self.train_targets)
        with self.model.cached_predict():
            self.model.predict(self.text_data_valid)
            self.model.predict(self.text_data_valid)

    def test_fields_too_long(self):
        """
        Smoke test for problems with multifield
        """
        self.model = MultiFieldClassifier()
        n_examples = 10
        self.model.fit([["text " * 100] * 10] * n_examples, ["target"] * n_examples)
        self.model.fit([["text", "text", "text", "text " * 300, "text " * 300]] * n_examples, ["target"] * n_examples)

    def test_multifield_regression(self):
        """                                                                                                                                                                         
        Ensure fit predict works with regression targets and multiple inputs.
        Ensure saving + loading does not cause errors                                                                                                                               
        Ensure saving + loading does not change predictions                                                                                                                         
        """
        self.model = MultiFieldRegressor()
        self.model.fit(self.text_data_train, [np.random.random() for _ in self.train_targets])
        predictions = self.model.predict(self.text_data_valid)
        self.model.save(self.save_file)
        model = MultiFieldRegressor.load(self.save_file)
        new_predictions = model.predict(self.text_data_valid)
        for new_pred, old_pred in zip(new_predictions, predictions):
            self.assertAlmostEqual(new_pred, old_pred, places=2)

    def test_regressor(self):
        n_samples = 20
        x_test = np.array(['the quick fox jumped over the lazy brown dog'] * n_samples)
        y_test = np.random.random(n_samples)
        model_test = Regressor(n_epochs=1)
        model_test.fit(x_test, y_test)
        preds = model_test.predict(x_test)
        self.assertIsInstance(preds, list)
        self.assertIsInstance(preds[0], list)
        self.assertIsInstance(preds[0][0], (float, np.float32))
